"""Tests for main module."""

import os
import sys
import time
from pathlib import Path
from unittest.mock import Mock, patch
from chad.__main__ import (
    main,
    _start_parent_watchdog,
    write_server_port,
    read_server_port,
    get_chad_dir,
)


class TestMain:
    """Test cases for main function."""

    @patch("chad.__main__.run_unified")
    @patch("chad.__main__.ConfigManager")
    def test_main_existing_user(self, mock_config_class, mock_run_unified):
        """Test main with existing user - password verified via ConfigManager."""
        mock_config = Mock()
        mock_config.is_first_run.return_value = False
        mock_config.verify_main_password.return_value = "verified-password"
        mock_config.get_cleanup_days.return_value = 3
        mock_config.get_ui_mode.return_value = "cli"
        mock_config_class.return_value = mock_config

        with patch.object(sys, "argv", ["chad"]):
            with patch.dict(os.environ, {}, clear=True):
                # Remove CHAD_PASSWORD to force interactive mode
                os.environ.pop("CHAD_PASSWORD", None)
                result = main()

        assert result == 0
        mock_config.verify_main_password.assert_called_once()
        mock_run_unified.assert_called_once()
        call_args = mock_run_unified.call_args
        assert call_args.args[0] == "verified-password"  # main_password

    @patch("chad.__main__.run_unified")
    @patch("chad.__main__.ConfigManager")
    def test_main_first_run(self, mock_config_class, mock_run_unified):
        """Test main with first run - prompts for password setup."""
        mock_config = Mock()
        mock_config.is_first_run.return_value = True
        mock_config.setup_main_password.return_value = "new-password"
        mock_config.get_cleanup_days.return_value = 3
        mock_config.get_ui_mode.return_value = "cli"
        mock_config_class.return_value = mock_config

        with patch.object(sys, "argv", ["chad"]):
            with patch.dict(os.environ, {}, clear=True):
                os.environ.pop("CHAD_PASSWORD", None)
                result = main()

        assert result == 0
        mock_config.setup_main_password.assert_called_once()
        mock_run_unified.assert_called_once()
        call_args = mock_run_unified.call_args
        assert call_args.args[0] == "new-password"  # main_password

    @patch("chad.__main__.run_unified")
    @patch("chad.__main__.ConfigManager")
    def test_main_with_env_password(self, mock_config_class, mock_run_unified):
        """Test main uses CHAD_PASSWORD env var when set."""
        mock_config = Mock()
        mock_config.is_first_run.return_value = False
        mock_config.get_cleanup_days.return_value = 3
        mock_config.get_ui_mode.return_value = "cli"
        mock_config_class.return_value = mock_config

        with patch.object(sys, "argv", ["chad"]):
            with patch.dict(os.environ, {"CHAD_PASSWORD": "env-password"}):
                result = main()

        assert result == 0
        # Should NOT call verify_main_password when env var is set
        mock_config.verify_main_password.assert_not_called()
        mock_config.setup_main_password.assert_not_called()
        mock_run_unified.assert_called_once()
        call_args = mock_run_unified.call_args
        assert call_args.args[0] == "env-password"

    @patch("chad.__main__.run_unified")
    @patch("chad.__main__.ConfigManager")
    def test_main_launch_error(self, mock_config_class, mock_run_unified):
        """Test main when UI launch fails."""
        mock_config = Mock()
        mock_config.is_first_run.return_value = False
        mock_config.verify_main_password.return_value = "password"
        mock_config.get_cleanup_days.return_value = 3
        mock_config.get_ui_mode.return_value = "cli"
        mock_config_class.return_value = mock_config

        mock_run_unified.side_effect = ValueError("Invalid password")

        with patch.object(sys, "argv", ["chad"]):
            with patch.dict(os.environ, {}, clear=True):
                os.environ.pop("CHAD_PASSWORD", None)
                result = main()

        assert result == 1

    @patch("chad.__main__.run_unified")
    @patch("chad.__main__.ConfigManager")
    def test_main_keyboard_interrupt(self, mock_config_class, mock_run_unified):
        """Test main with keyboard interrupt."""
        mock_config = Mock()
        mock_config.is_first_run.return_value = False
        mock_config.verify_main_password.return_value = "password"
        mock_config.get_cleanup_days.return_value = 3
        mock_config.get_ui_mode.return_value = "cli"
        mock_config_class.return_value = mock_config

        mock_run_unified.side_effect = KeyboardInterrupt()

        with patch.object(sys, "argv", ["chad"]):
            with patch.dict(os.environ, {}, clear=True):
                os.environ.pop("CHAD_PASSWORD", None)
                result = main()

        assert result == 0

    @patch("chad.__main__.run_server")
    @patch("chad.__main__.ConfigManager")
    def test_main_server_only_mode(self, mock_config_class, mock_run_server):
        """Test main in server-only mode (no password needed)."""
        mock_config = Mock()
        mock_config.get_cleanup_days.return_value = 3
        mock_config_class.return_value = mock_config

        with patch.object(sys, "argv", ["chad", "--mode", "server"]):
            result = main()

        assert result == 0
        mock_run_server.assert_called_once()
        # Should NOT prompt for password in server mode
        mock_config.verify_main_password.assert_not_called()
        mock_config.setup_main_password.assert_not_called()

    @patch("chad.__main__.run_unified")
    @patch("chad.__main__.ConfigManager")
    def test_main_cli_mode(self, mock_config_class, mock_run_unified):
        """Test main with CLI ui mode."""
        mock_config = Mock()
        mock_config.is_first_run.return_value = False
        mock_config.verify_main_password.return_value = "password"
        mock_config.get_cleanup_days.return_value = 3
        mock_config.get_ui_mode.return_value = "cli"
        mock_config_class.return_value = mock_config

        with patch.object(sys, "argv", ["chad", "--ui", "cli"]):
            with patch.dict(os.environ, {}, clear=True):
                os.environ.pop("CHAD_PASSWORD", None)
                result = main()

        assert result == 0
        mock_run_unified.assert_called_once()
        call_kwargs = mock_run_unified.call_args.kwargs
        assert call_kwargs.get("ui_mode") == "cli"

    @patch("chad.__main__.run_unified")
    @patch("chad.__main__.ConfigManager")
    def test_main_server_url_auto_discovers_port(self, mock_config_class, mock_run_unified, tmp_path):
        """Test that --server-url auto autodiscovers port from file."""
        mock_config = Mock()
        mock_config.is_first_run.return_value = False
        mock_config.verify_main_password.return_value = "password"
        mock_config.get_cleanup_days.return_value = 3
        mock_config.get_ui_mode.return_value = "cli"
        mock_config_class.return_value = mock_config

        # Write a port file for autodiscovery
        port_file = tmp_path / "server.port"
        port_file.write_text("9876\n")

        with patch.object(sys, "argv", ["chad", "--server-url", "auto"]):
            with patch.dict(os.environ, {"CHAD_DIR": str(tmp_path), "CHAD_PASSWORD": "test"}):
                result = main()

        assert result == 0
        mock_run_unified.assert_called_once()
        call_kwargs = mock_run_unified.call_args.kwargs
        assert call_kwargs.get("server_url") == "http://127.0.0.1:9876"

    @patch("chad.__main__.run_unified")
    @patch("chad.__main__.ConfigManager")
    def test_main_server_url_skips_password_prompt(self, mock_config_class, mock_run_unified):
        """Connecting to existing server should not ask for main password."""
        mock_config = Mock()
        mock_config.is_first_run.return_value = False
        mock_config.get_cleanup_days.return_value = 3
        mock_config.get_ui_mode.return_value = "cli"
        mock_config_class.return_value = mock_config

        with patch.object(sys, "argv", ["chad", "--server-url", "http://127.0.0.1:9999"]):
            with patch.dict(os.environ, {}, clear=True):
                os.environ.pop("CHAD_PASSWORD", None)
                result = main()

        assert result == 0
        mock_config.verify_main_password.assert_not_called()
        mock_config.setup_main_password.assert_not_called()
        mock_run_unified.assert_called_once()
        call_args = mock_run_unified.call_args
        assert call_args.args[0] is None  # main_password
        assert call_args.kwargs.get("server_url") == "http://127.0.0.1:9999"

    @patch("chad.__main__.run_unified")
    @patch("chad.__main__.ConfigManager")
    def test_main_server_url_auto_skips_password_prompt(self, mock_config_class, mock_run_unified, tmp_path):
        """Autodiscovery should also skip password prompts."""
        mock_config = Mock()
        mock_config.is_first_run.return_value = False
        mock_config.get_cleanup_days.return_value = 3
        mock_config.get_ui_mode.return_value = "cli"
        mock_config_class.return_value = mock_config

        port_file = tmp_path / "server.port"
        port_file.write_text("5555\n")

        with patch.object(sys, "argv", ["chad", "--server-url", "auto"]):
            with patch.dict(os.environ, {"CHAD_DIR": str(tmp_path)}, clear=True):
                result = main()

        assert result == 0
        mock_config.verify_main_password.assert_not_called()
        mock_config.setup_main_password.assert_not_called()
        mock_run_unified.assert_called_once()
        call_args = mock_run_unified.call_args
        assert call_args.args[0] is None
        assert call_args.kwargs.get("server_url") == "http://127.0.0.1:5555"

    @patch("chad.__main__.run_unified")
    @patch("chad.__main__.ConfigManager")
    def test_main_server_url_auto_fails_when_no_port_file(self, mock_config_class, mock_run_unified, tmp_path):
        """Test that --server-url auto fails gracefully when port file missing."""
        mock_config = Mock()
        mock_config.is_first_run.return_value = False
        mock_config.verify_main_password.return_value = "password"
        mock_config.get_cleanup_days.return_value = 3
        mock_config.get_ui_mode.return_value = "cli"
        mock_config_class.return_value = mock_config

        with patch.object(sys, "argv", ["chad", "--server-url", "auto"]):
            with patch.dict(os.environ, {"CHAD_DIR": str(tmp_path), "CHAD_PASSWORD": "test"}):
                result = main()

        assert result == 1
        mock_run_unified.assert_not_called()


class TestParentWatchdog:
    """Test cases for parent process watchdog."""

    def test_watchdog_not_started_without_env_var(self, monkeypatch):
        """Watchdog thread should not start without CHAD_PARENT_PID."""
        monkeypatch.delenv("CHAD_PARENT_PID", raising=False)

        thread = _start_parent_watchdog()
        assert thread is None

    def test_watchdog_not_started_with_invalid_pid(self, monkeypatch):
        """Watchdog thread should not start with invalid PID."""
        monkeypatch.setenv("CHAD_PARENT_PID", "not_a_number")

        thread = _start_parent_watchdog()
        assert thread is None

    def test_watchdog_starts_with_valid_pid(self, monkeypatch):
        """Watchdog thread should start with valid parent PID."""
        # Use current process as parent (it exists, so watchdog won't kill us)
        monkeypatch.setenv("CHAD_PARENT_PID", str(os.getpid()))

        thread = _start_parent_watchdog()
        assert thread is not None
        # Give thread time to start
        time.sleep(0.05)
        assert thread.is_alive()


class TestServerPortAutodiscovery:
    """Test cases for server port autodiscovery."""

    def test_write_server_port(self, tmp_path):
        """Test writing server port to file."""
        with patch.dict(os.environ, {"CHAD_DIR": str(tmp_path)}):
            write_server_port(8765)

            port_file = tmp_path / "server.port"
            assert port_file.exists()
            assert port_file.read_text().strip() == "8765"

    def test_read_server_port(self, tmp_path):
        """Test reading server port from file."""
        port_file = tmp_path / "server.port"
        port_file.write_text("9999\n")

        with patch.dict(os.environ, {"CHAD_DIR": str(tmp_path)}):
            port = read_server_port()

        assert port == 9999

    def test_read_server_port_returns_none_when_file_missing(self, tmp_path):
        """Test reading port returns None when file doesn't exist."""
        with patch.dict(os.environ, {"CHAD_DIR": str(tmp_path)}):
            port = read_server_port()

        assert port is None

    def test_read_server_port_returns_none_when_invalid(self, tmp_path):
        """Test reading port returns None when file has invalid content."""
        port_file = tmp_path / "server.port"
        port_file.write_text("not_a_number\n")

        with patch.dict(os.environ, {"CHAD_DIR": str(tmp_path)}):
            port = read_server_port()

        assert port is None

    def test_get_chad_dir_uses_env(self, tmp_path):
        """Test get_chad_dir uses CHAD_DIR env var when set."""
        with patch.dict(os.environ, {"CHAD_DIR": str(tmp_path)}):
            result = get_chad_dir()

        assert result == tmp_path

    def test_get_chad_dir_defaults_to_home(self):
        """Test get_chad_dir defaults to ~/.chad when CHAD_DIR not set."""
        with patch.dict(os.environ, {}, clear=False):
            # Remove CHAD_DIR if present
            os.environ.pop("CHAD_DIR", None)
            result = get_chad_dir()

        assert result == Path.home() / ".chad"



"""Tests for the reusable test utility tools in test_helpers.py."""

from __future__ import annotations

import json
import time

import pytest
from fastapi.testclient import TestClient

from chad.server.main import create_app
from chad.server.services import reset_session_manager, reset_task_executor
from chad.server.services.pty_stream import reset_pty_stream_service
from chad.server.state import reset_state

from tests.test_helpers import (
    CollectedEvents,
    CapturedCommand,
    ConfigParityResult,
    ProviderOutputSimulator,
    StreamInspection,
    TaskPhaseMonitor,
    capture_provider_command,
    cli_config_parity_check,
    collect_stream_events,
    inspect_stream_output,
)


# ── Fixtures ───────────────────────────────────────────────────────────────


@pytest.fixture
def client(tmp_path, monkeypatch):
    """Create an isolated FastAPI test client."""
    temp_config = tmp_path / "test_chad.conf"
    monkeypatch.setenv("CHAD_CONFIG", str(temp_config))
    monkeypatch.setenv("CHAD_LOG_DIR", str(tmp_path / "logs"))

    initial_config = {
        "encryption_salt": "dGVzdHNhbHQ=",
        "password_hash": "",
        "accounts": {},
    }
    temp_config.write_text(json.dumps(initial_config))

    reset_session_manager()
    reset_task_executor()
    reset_pty_stream_service()
    reset_state()

    app = create_app()
    with TestClient(app) as c:
        yield c

    reset_session_manager()
    reset_task_executor()
    reset_pty_stream_service()
    reset_state()


@pytest.fixture
def git_repo(tmp_path):
    """Create a temporary git repo."""
    import subprocess

    repo = tmp_path / "test_repo"
    repo.mkdir()
    subprocess.run(["git", "init"], cwd=repo, capture_output=True, check=True)
    subprocess.run(
        ["git", "config", "user.email", "test@test.com"],
        cwd=repo, capture_output=True, check=True,
    )
    subprocess.run(
        ["git", "config", "user.name", "Test"],
        cwd=repo, capture_output=True, check=True,
    )
    (repo / "README.md").write_text("# Test\n")
    (repo / "BUGS.md").write_text("# Bugs\n")
    subprocess.run(["git", "add", "."], cwd=repo, capture_output=True, check=True)
    subprocess.run(
        ["git", "commit", "-m", "init"],
        cwd=repo, capture_output=True, check=True,
    )
    return repo


def _create_session_and_account(client, git_repo, account_name="mock-agent"):
    """Helper: create a session and add a mock account via API."""
    sess = client.post("/api/v1/sessions", json={"name": "tool-test"})
    assert sess.status_code == 201
    session_id = sess.json()["id"]

    acc = client.post(
        "/api/v1/accounts",
        json={"name": account_name, "provider": "mock"},
    )
    assert acc.status_code in (200, 201, 409)

    return session_id


# ── Tool 1: collect_stream_events ─────────────────────────────────────────


class TestCollectStreamEvents:
    """Tests for collect_stream_events."""

    def test_collects_events_from_completed_task(self, client, git_repo):
        session_id = _create_session_and_account(client, git_repo)

        task = client.post(
            f"/api/v1/sessions/{session_id}/tasks",
            json={
                "project_path": str(git_repo),
                "task_description": "Test task",
                "coding_agent": "mock-agent",
            },
        )
        assert task.status_code in (200, 201)

        # Give mock agent time to run
        time.sleep(2)

        events = collect_stream_events(
            client, session_id, timeout=10, wait_for_completion=True
        )

        assert isinstance(events, CollectedEvents)
        assert len(events.all_events) > 0
        assert isinstance(events.terminal_events, list)
        assert isinstance(events.structured_events, list)
        assert isinstance(events.decoded_output, str)

    def test_returns_empty_on_no_events(self, client):
        sess = client.post("/api/v1/sessions", json={"name": "empty"})
        session_id = sess.json()["id"]

        events = collect_stream_events(
            client, session_id, timeout=1, wait_for_completion=False
        )

        assert isinstance(events, CollectedEvents)
        assert events.all_events == []
        assert events.decoded_output == ""

    def test_non_blocking_mode_returns_immediately(self, client, git_repo):
        session_id = _create_session_and_account(client, git_repo)

        start = time.monotonic()
        events = collect_stream_events(
            client, session_id, timeout=5, wait_for_completion=False
        )
        elapsed = time.monotonic() - start

        assert isinstance(events, CollectedEvents)
        # Should return quickly, not wait for timeout
        assert elapsed < 3


# ── Tool 2: ProviderOutputSimulator ───────────────────────────────────────


class TestProviderOutputSimulator:
    """Tests for ProviderOutputSimulator."""

    def test_unknown_scenario_raises(self, monkeypatch):
        with pytest.raises(ValueError, match="Unknown scenario"):
            ProviderOutputSimulator(monkeypatch, "nonexistent_scenario")

    def test_qwen_duplicate_scenario_patches(self, client, git_repo, monkeypatch):
        session_id = _create_session_and_account(client, git_repo)
        ProviderOutputSimulator(monkeypatch, "qwen_duplicate")

        task = client.post(
            f"/api/v1/sessions/{session_id}/tasks",
            json={
                "project_path": str(git_repo),
                "task_description": "Test qwen",
                "coding_agent": "mock-agent",
            },
        )
        assert task.status_code in (200, 201)

        time.sleep(3)
        events = collect_stream_events(client, session_id, timeout=5)
        # The simulated script should have produced some terminal output
        assert isinstance(events.decoded_output, str)

    def test_codex_system_prompt_scenario(self, client, git_repo, monkeypatch):
        session_id = _create_session_and_account(client, git_repo)
        ProviderOutputSimulator(monkeypatch, "codex_system_prompt")

        task = client.post(
            f"/api/v1/sessions/{session_id}/tasks",
            json={
                "project_path": str(git_repo),
                "task_description": "Test codex",
                "coding_agent": "mock-agent",
            },
        )
        assert task.status_code in (200, 201)
        time.sleep(3)

        events = collect_stream_events(client, session_id, timeout=5)
        assert isinstance(events, CollectedEvents)

    def test_all_scenarios_are_valid(self, monkeypatch):
        """Every registered scenario should be installable without error."""
        from tests.test_helpers import SCENARIOS

        for name in SCENARIOS:
            # Reinstantiate monkeypatch context each time
            ProviderOutputSimulator(monkeypatch, name)


# ── Tool 3: TaskPhaseMonitor ──────────────────────────────────────────────


class TestTaskPhaseMonitor:
    """Tests for TaskPhaseMonitor."""

    def test_empty_events(self):
        monitor = TaskPhaseMonitor([])
        assert monitor.phases == []
        assert monitor.phase_names() == []

    def test_coding_phase_detection(self):
        events = [
            {"type": "session_started", "seq": 1},
            {"type": "terminal_output", "seq": 2, "data": "Hello"},
            {"type": "terminal_output", "seq": 3, "data": "World"},
        ]
        monitor = TaskPhaseMonitor(events)
        assert "coding" in monitor.phase_names()
        counts = monitor.terminal_counts_by_phase()
        assert counts.get("coding", 0) == 2

    def test_verification_phase_detection(self):
        events = [
            {"type": "session_started", "seq": 1},
            {"type": "terminal_output", "seq": 2, "data": "code output"},
            {"type": "verification_attempt", "seq": 3, "attempt_number": 1},
            {"type": "terminal_output", "seq": 4, "data": "verify output"},
            {"type": "terminal_output", "seq": 5, "data": "more verify"},
        ]
        monitor = TaskPhaseMonitor(events)
        names = monitor.phase_names()
        assert "coding" in names
        assert "verification_1" in names
        counts = monitor.terminal_counts_by_phase()
        assert counts.get("coding", 0) == 1
        assert counts.get("verification_1", 0) == 2

    def test_multiple_verification_rounds(self):
        events = [
            {"type": "session_started", "seq": 1},
            {"type": "terminal_output", "seq": 2, "data": "coding"},
            {"type": "verification_attempt", "seq": 3, "attempt_number": 1},
            {"type": "terminal_output", "seq": 4, "data": "v1"},
            {"type": "verification_attempt", "seq": 5, "attempt_number": 2},
            {"type": "terminal_output", "seq": 6, "data": "v2"},
        ]
        monitor = TaskPhaseMonitor(events)
        names = monitor.phase_names()
        assert names == ["coding", "verification_1", "verification_2"]

    def test_text_based_phase_markers(self):
        events = [
            {"type": "terminal_output", "seq": 1, "data": "Phase 1: explore"},
            {"type": "terminal_output", "seq": 2, "data": "reading files"},
            {"type": "terminal_output", "seq": 3, "data": "Phase 2: implement"},
            {"type": "terminal_output", "seq": 4, "data": "writing code"},
        ]
        monitor = TaskPhaseMonitor(events)
        names = monitor.phase_names()
        assert "phase_1" in names
        assert "phase_2" in names


# ── Tool 4: capture_provider_command ──────────────────────────────────────


class TestCaptureProviderCommand:
    """Tests for capture_provider_command."""

    def test_mock_provider(self, tmp_path):
        result = capture_provider_command(
            provider="mock",
            account_name="test-mock",
            project_path=str(tmp_path),
            task_description="Fix the bug",
        )
        assert isinstance(result, CapturedCommand)
        assert isinstance(result.cmd, list)
        assert len(result.cmd) > 0
        assert result.cmd[0] == "python3"

    def test_anthropic_provider_includes_stream_json(self, tmp_path):
        result = capture_provider_command(
            provider="anthropic",
            account_name="my-claude",
            project_path=str(tmp_path),
            task_description="Test task",
        )
        assert "--output-format" in result.cmd
        assert "stream-json" in result.cmd
        assert "CLAUDE_CONFIG_DIR" in result.env

    def test_openai_provider_uses_exec(self, tmp_path):
        result = capture_provider_command(
            provider="openai",
            account_name="my-codex",
            project_path=str(tmp_path),
            task_description="Test task",
        )
        assert "exec" in result.cmd
        assert "HOME" in result.env
        # Codex reads prompt from stdin
        assert result.initial_input is not None

    def test_qwen_provider(self, tmp_path):
        result = capture_provider_command(
            provider="qwen",
            account_name="my-qwen",
            project_path=str(tmp_path),
            task_description="Test task",
        )
        assert "--output-format" in result.cmd
        assert "stream-json" in result.cmd

    def test_no_task_description(self, tmp_path):
        result = capture_provider_command(
            provider="mock",
            account_name="test",
            project_path=str(tmp_path),
        )
        assert isinstance(result, CapturedCommand)

    def test_implementation_phase(self, tmp_path):
        result = capture_provider_command(
            provider="anthropic",
            account_name="test",
            project_path=str(tmp_path),
            task_description="Implement feature",
            phase="implementation",
            exploration_output="Found 5 relevant files.",
        )
        assert isinstance(result.cmd, list)


# ── Tool 5: cli_config_parity_check ──────────────────────────────────────


class TestCliConfigParityCheck:
    """Tests for cli_config_parity_check."""

    def test_returns_result(self):
        result = cli_config_parity_check()
        assert isinstance(result, ConfigParityResult)
        assert isinstance(result.api_keys, set)
        assert isinstance(result.cli_keys, set)
        assert isinstance(result.missing_from_cli, set)

    def test_api_keys_excludes_internal(self):
        result = cli_config_parity_check()
        internal = {
            "password_hash", "encryption_salt", "accounts",
            "role_assignments", "preferences", "projects",
            "mock_remaining_usage",
        }
        assert not result.api_keys.intersection(internal)

    def test_coverage_is_union(self):
        result = cli_config_parity_check()
        assert result.cli_keys | result.missing_from_cli == result.api_keys


# ── Tool 6: inspect_stream_output ─────────────────────────────────────────


class TestInspectStreamOutput:
    """Tests for inspect_stream_output."""

    def test_clean_output(self):
        result = inspect_stream_output("Hello world\nTask complete")
        assert isinstance(result, StreamInspection)
        assert not result.has_raw_json
        assert not result.has_binary_data
        assert result.json_fragments == []
        assert result.binary_fragments == []

    def test_detects_raw_json(self):
        output = 'Some text {"type": "assistant", "message": {"content": "hi"}} more text'
        result = inspect_stream_output(output)
        assert result.has_raw_json
        assert len(result.json_fragments) > 0

    def test_detects_binary_garbage(self):
        output = "Normal text\n@@@@@@@%#%%#@@@%#####%@@%%%%#%%%%\nMore text"
        result = inspect_stream_output(output)
        assert result.has_binary_data
        assert len(result.binary_fragments) > 0

    def test_detects_both(self):
        output = '{"type": "item"} and @@@@@@@@@@@@@@@@@@'
        result = inspect_stream_output(output)
        assert result.has_raw_json
        assert result.has_binary_data

    def test_content_pattern_detection(self):
        output = '{"content": "some value"}'
        result = inspect_stream_output(output)
        assert result.has_raw_json

    def test_short_at_signs_not_flagged(self):
        # Less than 10 consecutive chars should not be flagged
        result = inspect_stream_output("email@user.com @@merge")
        assert not result.has_binary_data

    def test_message_pattern(self):
        output = '{"message": "hello"}'
        result = inspect_stream_output(output)
        assert result.has_raw_json

    def test_role_pattern(self):
        output = '{"role": "assistant"}'
        result = inspect_stream_output(output)
        assert result.has_raw_json



"""Services layer for Chad server."""

from .session_manager import (
    Session,
    SessionManager,
    get_session_manager,
    reset_session_manager,
)
from .task_executor import (
    Task,
    TaskState,
    TaskExecutor,
    StreamEvent,
    get_task_executor,
    reset_task_executor,
)

__all__ = [
    # Session manager
    "Session",
    "SessionManager",
    "get_session_manager",
    "reset_session_manager",
    # Task executor
    "Task",
    "TaskState",
    "TaskExecutor",
    "StreamEvent",
    "get_task_executor",
    "reset_task_executor",
]



"""Status endpoint."""

import os

from fastapi import APIRouter
from pydantic import BaseModel, Field

from chad.server import __version__
from chad.server.state import get_uptime

router = APIRouter()

_server_cwd: str = os.getcwd()


class StatusResponse(BaseModel):
    """Response model for status endpoint."""

    status: str = Field(default="healthy", description="Server status")
    version: str = Field(description="Server version")
    uptime_seconds: float = Field(description="Server uptime in seconds")
    cwd: str = Field(description="Server working directory")


@router.get("/status", response_model=StatusResponse)
async def get_status() -> StatusResponse:
    """Status endpoint reporting server health, version, and uptime."""
    return StatusResponse(
        status="healthy",
        version=__version__,
        uptime_seconds=get_uptime(),
        cwd=os.environ.get("CHAD_PROJECT_PATH", _server_cwd),
    )



"""Tests for Codex prompt echo filtering."""

import re


def strip_ansi(text: str) -> str:
    """Remove ANSI escape codes from text."""
    ansi_pattern = re.compile(r'\x1b\[[0-9;]*m')
    return ansi_pattern.sub('', text)


# Simulated Codex output with ANSI codes
CODEX_RAW_OUTPUT = """\
\x1b[2m2026-02-05T21:42:38.165352Z\x1b[0m \x1b[31mERROR\x1b[0m some error message
OpenAI Codex v0.92.0 (research preview)
--------
\x1b[1mworkdir:\x1b[0m /tmp/tmpjwlejti6
\x1b[1mmodel:\x1b[0m gpt-5.2-codex
\x1b[1mprovider:\x1b[0m openai
\x1b[1mapproval:\x1b[0m never
\x1b[1msandbox:\x1b[0m danger-full-access
\x1b[1msession id:\x1b[0m 019c2fc1-d7a7-7d13-ac6a-0229bdab82ec
--------
\x1b[36muser\x1b[0m

## Verification

This is the prompt that should be filtered out.
It contains the user's task description.

\x1b[36mmcp startup:\x1b[0m no servers
\x1b[1mthinking\x1b[0m
**Agent is now working on the task**
This content should be shown to the user.
"""


def filter_codex_output(raw_output: str) -> str:
    """Filter Codex prompt echo from output.

    Returns the content that should be shown to the user.
    """
    normalized = raw_output.replace("\r\n", "\n").replace("\r", "\n")

    # Strip ANSI codes for pattern matching
    ansi_pattern = re.compile(r'\x1b\[[0-9;]*m')
    stripped = ansi_pattern.sub('', normalized)

    # Find the "user" marker (after second --------)
    user_line_match = re.search(r'\n--------\nuser\n', stripped)

    if not user_line_match:
        # No prompt echo detected, return as-is
        return raw_output

    # Find the position in original string
    user_pattern = re.compile(r'\n--------\n(?:\x1b\[[0-9;]*m)*user(?:\x1b\[[0-9;]*m)*\n')
    match = user_pattern.search(normalized)

    if not match:
        return raw_output

    # Content before the "user" line (banner + header)
    pre_echo = normalized[:match.start()]

    # Content after the "user" line (need to find mcp startup)
    post_user = normalized[match.end():]
    stripped_post = ansi_pattern.sub('', post_user)

    if "mcp startup:" not in stripped_post.lower():
        # No end marker yet, can't filter
        return raw_output

    # Find mcp startup marker
    mcp_pattern = re.compile(r'(?:\x1b\[[0-9;]*m)*mcp startup:(?:\x1b\[[0-9;]*m)*[^\n]*\n', re.IGNORECASE)
    mcp_match = mcp_pattern.search(post_user)

    if mcp_match:
        agent_output = post_user[mcp_match.end():]
    else:
        # Fallback
        marker_pos = stripped_post.lower().find("mcp startup:")
        newline_after = stripped_post.find("\n", marker_pos)
        if newline_after != -1:
            agent_output = post_user[newline_after + 1:]
        else:
            agent_output = ""

    # Return pre-echo (header) + agent output
    return pre_echo + "\n" + agent_output


class TestCodexFiltering:
    """Tests for Codex output filtering."""

    def test_filter_removes_prompt_echo(self):
        """Filter should remove the echoed prompt."""
        result = filter_codex_output(CODEX_RAW_OUTPUT)

        # Should NOT contain the prompt text
        assert "This is the prompt that should be filtered out" not in result
        assert "It contains the user's task description" not in result
        assert "## Verification" not in result

    def test_filter_keeps_banner_and_header(self):
        """Filter should keep the Codex banner and session info."""
        result = filter_codex_output(CODEX_RAW_OUTPUT)

        # Should contain banner
        assert "OpenAI Codex v0.92.0" in result

        # Should contain header info
        stripped = strip_ansi(result)
        assert "workdir:" in stripped
        assert "model:" in stripped
        assert "session id:" in stripped

    def test_filter_keeps_agent_output(self):
        """Filter should keep the agent's actual work output."""
        result = filter_codex_output(CODEX_RAW_OUTPUT)

        # Should contain agent work
        assert "Agent is now working on the task" in result
        assert "This content should be shown to the user" in result

    def test_filter_handles_no_prompt_echo(self):
        """Filter should pass through output without prompt echo."""
        simple_output = "Just some output\nNo markers here\n"
        result = filter_codex_output(simple_output)
        assert result == simple_output

    def test_filter_handles_ansi_codes_in_markers(self):
        """Filter should handle ANSI codes in the user/mcp markers."""
        # The actual output has ANSI codes like [36muser[0m
        result = filter_codex_output(CODEX_RAW_OUTPUT)

        # Verify we got to the agent output
        assert "thinking" in strip_ansi(result) or "Agent is now working" in result


class TestCodexBufferFlush:
    """Tests for Codex output buffer flush behavior."""

    def test_unflushed_buffer_loses_content(self):
        """Demonstrate that buffered content without mcp marker needs explicit flush."""
        # Simulate incremental Codex output that hasn't reached mcp startup marker
        buffer = ""
        captured = []

        # First chunk: header starts
        buffer += "OpenAI Codex v0.92.0\n--------\nworkdir: /tmp\n"
        # Buffer hasn't seen markers yet, nothing emitted

        # Stream ends unexpectedly - buffer has content
        assert buffer.strip()
        assert len(captured) == 0

        # Without flush, this content would be lost
        # With flush: captured_output.append(codex_output_buffer)
        captured.append(buffer)
        assert "OpenAI Codex" in "\n".join(captured)

    def test_full_output_not_truncated_by_terminal(self):
        """Accumulated text captures full output regardless of terminal screen size."""
        accumulated_text = []

        # Simulate 200 lines of agent output being accumulated
        for i in range(200):
            chunk = f"Agent output line {i}\n"
            accumulated_text.append(chunk)

        full_output = "\n".join(accumulated_text)

        # All lines should be present (unlike terminal.get_text() which only has last ~50)
        assert "Agent output line 0" in full_output
        assert "Agent output line 100" in full_output
        assert "Agent output line 199" in full_output


if __name__ == "__main__":
    import pytest
    pytest.main([__file__, "-v"])



"""WebSocket streaming Pydantic schemas."""

from typing import Any, Literal
from pydantic import BaseModel, Field


StreamMessageType = Literal[
    "stream",            # Raw AI output chunk
    "activity",          # Activity update (tool use, thinking)
    "status",            # Status message
    "message_start",     # AI message started
    "message_complete",  # AI message completed
    "progress",          # Progress update from agent
    "complete",          # Task completed
    "error",             # Error occurred
]


class StreamMessage(BaseModel):
    """WebSocket message format for streaming updates."""

    type: StreamMessageType = Field(description="Message type")
    session_id: str = Field(description="Session this message belongs to")
    data: dict[str, Any] = Field(default_factory=dict, description="Message payload")


class StreamChunkData(BaseModel):
    """Data payload for stream chunks."""

    chunk: str = Field(description="Text chunk from AI")


class ActivityData(BaseModel):
    """Data payload for activity updates."""

    activity_type: str = Field(description="Type of activity (tool, thinking, etc.)")
    detail: str = Field(description="Activity detail/description")


class StatusData(BaseModel):
    """Data payload for status updates."""

    status: str = Field(description="Status message")


class MessageStartData(BaseModel):
    """Data payload for message start."""

    speaker: str = Field(description="Who is speaking (CODING AI, VERIFICATION AI, etc.)")


class MessageCompleteData(BaseModel):
    """Data payload for message complete."""

    speaker: str = Field(description="Who spoke")
    content: str = Field(description="Full message content")


class ProgressData(BaseModel):
    """Data payload for progress updates."""

    summary: str = Field(description="Progress summary")
    location: str | None = Field(default=None, description="Code location")
    next_step: str | None = Field(default=None, description="What the agent plans to do next")
    before_screenshot: str | None = Field(default=None, description="Path to before screenshot")
    before_description: str | None = Field(default=None, description="Screenshot description")


class CompleteData(BaseModel):
    """Data payload for task completion."""

    success: bool = Field(description="Whether task succeeded")
    message: str = Field(description="Completion message")
    change_summary: str | None = Field(default=None, description="Summary of changes made")


class ErrorData(BaseModel):
    """Data payload for errors."""

    error: str = Field(description="Error message")
    details: str | None = Field(default=None, description="Additional error details")


# Client -> Server messages
class ClientMessage(BaseModel):
    """Message from client to server via WebSocket."""

    type: Literal["subscribe", "cancel", "ping"] = Field(description="Client message type")


class SubscribeMessage(ClientMessage):
    """Client message to subscribe to session updates."""

    type: Literal["subscribe"] = "subscribe"


class CancelMessage(ClientMessage):
    """Client message to request task cancellation."""

    type: Literal["cancel"] = "cancel"


class PingMessage(ClientMessage):
    """Client ping message for keepalive."""

    type: Literal["ping"] = "ping"



"""Structured event logging for session handovers.

Events are stored as JSONL (one JSON object per line) in ~/.chad/logs/{session_id}.jsonl
Large artifacts (stdout/stderr >10KB) are stored separately in ~/.chad/logs/artifacts/
"""

from __future__ import annotations

import hashlib
import json
import os
import uuid
from dataclasses import asdict, dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Literal


# Event types
EventType = Literal[
    "session_started",
    "status",
    "progress",
    "model_selected",
    "provider_switched",
    "user_message",
    "assistant_message",
    "tool_declared",
    "tool_call_started",
    "tool_call_finished",
    "verification_attempt",
    "context_condensed",
    "terminal_output",
    "milestone",
    "session_ended",
]


@dataclass
class ArtifactRef:
    """Reference to an artifact file."""

    path: str  # Relative path from logs directory
    sha256: str
    size: int


@dataclass
class MessageBlock:
    """A block within an assistant message."""

    kind: Literal["text", "thinking", "tool_call", "tool_result", "error"]
    content: str = ""
    tool: str | None = None
    tool_call_id: str | None = None
    args: dict[str, Any] | None = None


@dataclass
class EventBase:
    """Base class for all events."""

    event_id: str = field(default_factory=lambda: str(uuid.uuid4()))
    ts: str = field(default_factory=lambda: datetime.now(timezone.utc).isoformat())
    seq: int = 0
    session_id: str = ""
    turn_id: str | None = None

    def to_dict(self) -> dict[str, Any]:
        """Convert event to dictionary for JSON serialization."""
        d = asdict(self)
        d["type"] = self.event_type
        return d

    @property
    def event_type(self) -> str:
        """Return the event type name."""
        # Convert class name to snake_case
        name = self.__class__.__name__
        if name.endswith("Event"):
            name = name[:-5]
        # Convert CamelCase to snake_case
        result = []
        for i, c in enumerate(name):
            if c.isupper() and i > 0:
                result.append("_")
            result.append(c.lower())
        return "".join(result)


@dataclass
class SessionStartedEvent(EventBase):
    """Logged when a session starts."""

    task_description: str = ""
    project_path: str = ""
    coding_provider: str = ""
    coding_account: str = ""
    coding_model: str | None = None


@dataclass
class StatusEvent(EventBase):
    """Logged when task execution status changes."""

    status: str = ""


@dataclass
class ProgressEvent(EventBase):
    """Logged when exploration emits a structured progress update."""

    summary: str = ""
    location: str | None = None
    next_step: str | None = None
    before_screenshot: str | None = None
    before_description: str | None = None


@dataclass
class ModelSelectedEvent(EventBase):
    """Logged when a model is selected or changed."""

    provider: str = ""
    model: str = ""
    reasoning_effort: str | None = None


@dataclass
class ProviderSwitchedEvent(EventBase):
    """Logged when switching between providers."""

    from_provider: str = ""
    to_provider: str = ""
    from_model: str = ""
    to_model: str = ""
    reason: str = ""


@dataclass
class UserMessageEvent(EventBase):
    """Logged when user sends a message."""

    content: str = ""


@dataclass
class AssistantMessageEvent(EventBase):
    """Logged when assistant responds."""

    blocks: list[dict[str, Any]] = field(default_factory=list)

    @classmethod
    def from_blocks(cls, blocks: list[MessageBlock], **kwargs) -> "AssistantMessageEvent":
        """Create from MessageBlock objects."""
        return cls(blocks=[asdict(b) for b in blocks], **kwargs)


@dataclass
class ToolDeclaredEvent(EventBase):
    """Logged when a tool is declared/available."""

    name: str = ""
    args_schema: dict[str, Any] = field(default_factory=dict)
    version: str = "1.0"


@dataclass
class ToolCallStartedEvent(EventBase):
    """Logged when a tool call begins."""

    tool_call_id: str = field(default_factory=lambda: f"tc_{uuid.uuid4().hex[:8]}")
    tool: str = ""  # "bash", "read", "write", "edit", "mcp", "glob", "grep"

    # For bash commands
    cwd: str | None = None
    command: str | None = None
    env_redactions: list[str] | None = None
    timeout_s: float | None = None

    # For file operations
    path: str | None = None
    file_bytes: int | None = None
    sha256: str | None = None

    # For file edits
    before_sha256: str | None = None

    # For MCP tools
    server: str | None = None
    tool_name: str | None = None
    args: dict[str, Any] | None = None


@dataclass
class ToolCallFinishedEvent(EventBase):
    """Logged when a tool call completes."""

    tool_call_id: str = ""
    exit_code: int | None = None
    duration_ms: int = 0
    stdout_ref: dict[str, Any] | None = None  # ArtifactRef as dict
    stderr_ref: dict[str, Any] | None = None  # ArtifactRef as dict
    llm_summary: str = ""  # Bounded summary for handover

    # For file edits
    after_sha256: str | None = None
    patch_ref: dict[str, Any] | None = None  # ArtifactRef as dict


@dataclass
class VerificationAttemptEvent(EventBase):
    """Logged for each verification attempt."""

    attempt_number: int = 1
    tool_call_refs: list[str] = field(default_factory=list)
    passed: bool = False
    summary: str = ""
    issues: list[str] = field(default_factory=list)


@dataclass
class ContextCondensedEvent(EventBase):
    """Logged when context is condensed/summarized.

    Used for both context window management and provider handoffs.
    When policy="provider_handoff", the structured fields contain
    progress data for resuming on a different provider.
    """

    replaces_seq_range: tuple[int, int] = (0, 0)
    summary_text: str = ""
    policy: str = "rolling_window"
    # Handoff context (populated when policy="provider_handoff")
    original_task: str = ""
    files_changed: list[str] = field(default_factory=list)
    files_created: list[str] = field(default_factory=list)
    key_commands: list[str] = field(default_factory=list)
    remaining_work: str = ""
    provider_session_id: str | None = None  # For native resume (thread_id, session_id)


@dataclass
class TerminalOutputEvent(EventBase):
    """Logged for terminal screen content.

    Contains human-readable text extracted from the terminal screen,
    with ANSI sequences processed by the terminal emulator. Only logged
    when screen content meaningfully changes.
    """

    data: str = ""  # Human-readable screen text (processed by terminal emulator)


@dataclass
class MilestoneEvent(EventBase):
    """Logged when a task reaches a significant milestone."""

    milestone_type: str = ""
    # Types: exploration, coding_complete, verification_started,
    # verification_passed, verification_failed, revision_started,
    # session_limit_reached, weekly_limit_reached, usage_threshold
    title: str = ""  # Display-ready title for UIs (e.g. "Discovery", "Session Limit")
    summary: str = ""
    details: dict[str, Any] = field(default_factory=dict)


@dataclass
class SessionEndedEvent(EventBase):
    """Logged when a session ends."""

    success: bool = False
    reason: str = ""
    total_tool_calls: int = 0
    total_turns: int = 0


# Size threshold for storing artifacts separately (10KB)
ARTIFACT_SIZE_THRESHOLD = 10 * 1024

# Maximum artifact size (10MB)
MAX_ARTIFACT_SIZE = 10 * 1024 * 1024


class EventLog:
    """Manages structured event logging for a session."""

    def __init__(
        self,
        session_id: str,
        base_dir: Path | None = None,
    ):
        self.session_id = session_id
        self._seq = 0
        self._current_turn_id: str | None = None

        # Determine base directory
        env_dir = os.environ.get("CHAD_LOG_DIR")
        if base_dir:
            self.base_dir = Path(base_dir)
        elif env_dir:
            self.base_dir = Path(env_dir)
        else:
            self.base_dir = Path.home() / ".chad" / "logs"

        self.base_dir.mkdir(parents=True, exist_ok=True)

        # Artifacts directory
        self.artifacts_dir = self.base_dir / "artifacts" / session_id
        self.artifacts_dir.mkdir(parents=True, exist_ok=True)

        # Log file path
        self.log_path = self.base_dir / f"{session_id}.jsonl"

        # Seed sequence counter from existing log if present
        if self.log_path.exists():
            try:
                last_line = ""
                with open(self.log_path, "r", encoding="utf-8") as f:
                    for line in f:
                        if line.strip():
                            last_line = line
                if last_line:
                    last_event = json.loads(last_line)
                    self._seq = int(last_event.get("seq", 0))
            except Exception:
                # If log is unreadable, fall back to starting at 0
                self._seq = 0

    def _next_seq(self) -> int:
        """Get next sequence number."""
        self._seq += 1
        return self._seq

    def start_turn(self) -> str:
        """Start a new conversation turn, returns turn_id."""
        self._current_turn_id = str(uuid.uuid4())[:8]
        return self._current_turn_id

    def log(self, event: EventBase) -> None:
        """Log an event to the session log."""
        # Set sequence and session info
        event.seq = self._next_seq()
        event.session_id = self.session_id
        if event.turn_id is None:
            event.turn_id = self._current_turn_id

        # Serialize and append
        event_dict = event.to_dict()

        with open(self.log_path, "a", encoding="utf-8") as f:
            f.write(json.dumps(event_dict) + "\n")

    def store_artifact(
        self,
        content: bytes | str,
        name: str,
    ) -> ArtifactRef | None:
        """Store content as an artifact if it exceeds threshold.

        Args:
            content: The content to store
            name: Base name for the artifact file

        Returns:
            ArtifactRef if stored as artifact, None if content is small
        """
        if isinstance(content, str):
            content = content.encode("utf-8")

        size = len(content)

        # Truncate if too large
        if size > MAX_ARTIFACT_SIZE:
            content = content[:MAX_ARTIFACT_SIZE] + b"\n[TRUNCATED - exceeded 10MB limit]"
            size = len(content)

        if size < ARTIFACT_SIZE_THRESHOLD:
            return None

        # Calculate hash
        sha256 = hashlib.sha256(content).hexdigest()

        # Store file
        artifact_path = self.artifacts_dir / f"{name}_{sha256[:8]}.txt"
        with open(artifact_path, "wb") as f:
            f.write(content)

        # Return relative path
        rel_path = str(artifact_path.relative_to(self.base_dir))
        return ArtifactRef(path=rel_path, sha256=sha256, size=size)

    def get_events(
        self,
        since_seq: int = 0,
        event_types: list[str] | None = None,
    ) -> list[dict[str, Any]]:
        """Read events from the log.

        Args:
            since_seq: Return events after this sequence number
            event_types: Filter to these event types (None = all)

        Returns:
            List of event dictionaries
        """
        if not self.log_path.exists():
            return []

        events = []
        with open(self.log_path, encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if not line:
                    continue
                try:
                    event = json.loads(line)
                    if event.get("seq", 0) > since_seq:
                        if event_types is None or event.get("type") in event_types:
                            events.append(event)
                except json.JSONDecodeError:
                    continue

        return events

    def get_artifact(self, ref: ArtifactRef | dict[str, Any]) -> bytes | None:
        """Read an artifact's content.

        Args:
            ref: ArtifactRef or dict with path key

        Returns:
            Artifact content as bytes, or None if not found
        """
        if isinstance(ref, dict):
            path = ref.get("path", "")
        else:
            path = ref.path

        artifact_path = self.base_dir / path
        if not artifact_path.exists():
            return None

        with open(artifact_path, "rb") as f:
            return f.read()

    def get_latest_seq(self) -> int:
        """Get the latest sequence number in the log."""
        return self._seq

    def close(self) -> None:
        """Close the event log (cleanup if needed)."""
        pass  # Currently no cleanup needed

    @classmethod
    def get_log_dir(cls, base_dir: Path | None = None) -> Path:
        """Get the log directory path.

        Args:
            base_dir: Override base directory

        Returns:
            Path to the logs directory
        """
        if base_dir is None:
            env_dir = os.environ.get("CHAD_LOG_DIR")
            if env_dir:
                return Path(env_dir)
            else:
                return Path.home() / ".chad" / "logs"
        return base_dir

    @classmethod
    def list_sessions(cls, base_dir: Path | None = None) -> list[str]:
        """List all session IDs with logs.

        Args:
            base_dir: Override base directory

        Returns:
            List of session IDs
        """
        base_dir = cls.get_log_dir(base_dir)

        if not base_dir.exists():
            return []

        sessions = []
        for f in base_dir.glob("*.jsonl"):
            sessions.append(f.stem)

        return sorted(sessions)


def compute_file_sha256(path: Path) -> str:
    """Compute SHA256 hash of a file."""
    h = hashlib.sha256()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(8192), b""):
            h.update(chunk)
    return h.hexdigest()



"""Tests for Chad server API endpoints."""

from pathlib import Path

import pytest
from fastapi.testclient import TestClient

from chad.server.main import create_app
from chad.server.services import reset_session_manager, reset_task_executor
from chad.server.state import get_config_manager, reset_state


@pytest.fixture
def client(tmp_path, monkeypatch):
    """Create a test client for the API with isolated config."""
    # Use temporary config file to avoid reading real accounts
    temp_config = tmp_path / "test_chad.conf"
    monkeypatch.setenv("CHAD_CONFIG", str(temp_config))

    # Reset all state before each test
    reset_session_manager()
    reset_task_executor()
    reset_state()

    app = create_app()
    with TestClient(app) as client:
        yield client

    # Reset after test
    reset_session_manager()
    reset_task_executor()
    reset_state()


class TestStatusEndpoint:
    """Tests for status endpoint."""

    def test_status(self, client):
        """Status endpoint returns health, version, and uptime."""
        response = client.get("/status")
        assert response.status_code == 200
        data = response.json()
        assert data["status"] == "healthy"
        assert "version" in data
        assert "uptime_seconds" in data


class TestSessionEndpoints:
    """Tests for session management endpoints."""

    def test_create_session(self, client):
        """Can create a new session."""
        response = client.post("/api/v1/sessions", json={"name": "Test Session"})
        assert response.status_code == 201
        data = response.json()
        assert "id" in data
        assert data["name"] == "Test Session"
        assert data["active"] is False
        assert data["paused"] is False  # New paused field

    def test_create_session_with_project_path(self, client):
        """Can create session with project path."""
        response = client.post(
            "/api/v1/sessions",
            json={"name": "Test", "project_path": "/tmp/test-project"},
        )
        assert response.status_code == 201
        data = response.json()
        assert data["project_path"] == "/tmp/test-project"

    def test_list_sessions(self, client):
        """Can list all sessions."""
        # Create a session first
        client.post("/api/v1/sessions", json={"name": "Session 1"})
        client.post("/api/v1/sessions", json={"name": "Session 2"})

        response = client.get("/api/v1/sessions")
        assert response.status_code == 200
        data = response.json()
        assert data["total"] == 2
        assert len(data["sessions"]) == 2

    def test_get_session(self, client):
        """Can get a specific session."""
        # Create a session
        create_resp = client.post("/api/v1/sessions", json={"name": "Test"})
        session_id = create_resp.json()["id"]

        response = client.get(f"/api/v1/sessions/{session_id}")
        assert response.status_code == 200
        data = response.json()
        assert data["id"] == session_id

    def test_get_session_not_found(self, client):
        """Returns 404 for non-existent session."""
        response = client.get("/api/v1/sessions/nonexistent")
        assert response.status_code == 404

    def test_delete_session(self, client):
        """Can delete a session."""
        # Create a session
        create_resp = client.post("/api/v1/sessions", json={"name": "Test"})
        session_id = create_resp.json()["id"]

        # Delete it
        response = client.delete(f"/api/v1/sessions/{session_id}")
        assert response.status_code == 204

        # Verify it's gone
        response = client.get(f"/api/v1/sessions/{session_id}")
        assert response.status_code == 404

    def test_resume_session_not_paused(self, client):
        """Resume endpoint returns resumed=False if session is not paused."""
        # Create a session
        create_resp = client.post("/api/v1/sessions", json={"name": "Test"})
        session_id = create_resp.json()["id"]

        # Try to resume when not paused
        response = client.post(f"/api/v1/sessions/{session_id}/resume")
        assert response.status_code == 200
        data = response.json()
        assert data["session_id"] == session_id
        assert data["resumed"] is False
        assert "not paused" in data["message"]

    def test_resume_session_not_found(self, client):
        """Resume endpoint returns 404 for non-existent session."""
        response = client.post("/api/v1/sessions/nonexistent/resume")
        assert response.status_code == 404


class TestProviderEndpoints:
    """Tests for provider management endpoints."""

    def test_list_providers(self, client):
        """Can list all supported providers."""
        response = client.get("/api/v1/providers")
        assert response.status_code == 200
        data = response.json()
        providers = data["providers"]
        assert len(providers) >= 5  # anthropic, openai, gemini, qwen, mistral

        provider_types = [p["type"] for p in providers]
        assert "anthropic" in provider_types
        assert "openai" in provider_types
        assert "gemini" in provider_types
        assert "opencode" in provider_types
        assert "kimi" in provider_types

    @pytest.mark.parametrize("provider", ["opencode", "kimi"])
    def test_create_account_accepts_new_provider_types(self, client, provider):
        """Account create API should accept all provider types exposed in the setup UI."""
        config_mgr = get_config_manager()
        config_mgr.save_config(
            {
                "password_hash": "",
                "encryption_salt": "dGVzdHNhbHQ=",
                "accounts": {},
            }
        )

        response = client.post(
            "/api/v1/accounts",
            json={"name": f"{provider}-test", "provider": provider},
        )
        assert response.status_code == 201, response.text
        data = response.json()
        assert data["name"] == f"{provider}-test"
        assert data["provider"] == provider

    def test_list_accounts_empty(self, client):
        """List accounts returns empty list when no accounts configured."""
        response = client.get("/api/v1/accounts")
        assert response.status_code == 200
        data = response.json()
        assert data["total"] == 0
        assert data["accounts"] == []


class TestConfigEndpoints:
    """Tests for config management endpoints."""

    def test_get_cleanup_settings(self, client):
        """Can get cleanup settings."""
        response = client.get("/api/v1/config/cleanup")
        assert response.status_code == 200
        data = response.json()
        assert "cleanup_days" in data
        assert data["cleanup_days"] >= 1

    def test_get_preferences(self, client):
        """Can get user preferences."""
        response = client.get("/api/v1/config/preferences")
        assert response.status_code == 200
        data = response.json()
        assert "ui_mode" in data

    def test_get_verification_settings(self, client):
        """Can get verification settings."""
        response = client.get("/api/v1/config/verification")
        assert response.status_code == 200
        data = response.json()
        assert "enabled" in data
        assert "auto_run" in data

    def test_get_verification_agent_default(self, client):
        """Returns None when no verification agent is set."""
        response = client.get("/api/v1/config/verification-agent")
        assert response.status_code == 200
        data = response.json()
        assert "account_name" in data
        # Default is None (same as coding agent)
        assert data["account_name"] is None

    def test_set_verification_agent_none_marker(self, client):
        """Setting verification agent to VERIFICATION_NONE marker persists correctly."""
        # Set to the special marker value
        response = client.put(
            "/api/v1/config/verification-agent",
            json={"account_name": "__verification_none__"},
        )
        assert response.status_code == 200
        data = response.json()
        assert data["account_name"] == "__verification_none__"

        # Verify it persists when we get it back
        get_response = client.get("/api/v1/config/verification-agent")
        assert get_response.status_code == 200
        get_data = get_response.json()
        assert get_data["account_name"] == "__verification_none__"

    def test_set_verification_agent_none_clears(self, client):
        """Setting verification agent to None clears the setting."""
        # First set to the marker
        client.put(
            "/api/v1/config/verification-agent",
            json={"account_name": "__verification_none__"},
        )

        # Then clear by setting to None
        response = client.put(
            "/api/v1/config/verification-agent",
            json={"account_name": None},
        )
        assert response.status_code == 200
        data = response.json()
        assert data["account_name"] is None

        # Verify it's cleared
        get_response = client.get("/api/v1/config/verification-agent")
        assert get_response.status_code == 200
        get_data = get_response.json()
        assert get_data["account_name"] is None

    def test_update_verification_settings_partial(self, client):
        """Can partially update verification settings and disable verification."""
        # Disable verification only
        resp = client.put("/api/v1/config/verification", json={"enabled": False})
        assert resp.status_code == 200
        data = resp.json()
        assert data["enabled"] is False
        # auto_run should remain default True
        assert data["auto_run"] is True

        # Disable auto_run while enabled already false
        resp2 = client.put("/api/v1/config/verification", json={"auto_run": False})
        assert resp2.status_code == 200
        data2 = resp2.json()
        assert data2["enabled"] is False
        assert data2["auto_run"] is False

        # GET should reflect latest values
        resp3 = client.get("/api/v1/config/verification")
        assert resp3.status_code == 200
        data3 = resp3.json()
        assert data3["enabled"] is False
        assert data3["auto_run"] is False

    def test_set_action_settings_invalid_account_returns_400(self, client):
        """Action settings with invalid switch target should return 400."""
        response = client.put(
            "/api/v1/config/action-settings",
            json={"settings": [
                {"event": "session_usage", "threshold": 90, "action": "switch_provider", "target_account": "nonexistent"},
            ]},
        )

        assert response.status_code == 400
        assert "valid target_account" in response.json()["detail"]

    def test_mock_run_duration_endpoints(self, client):
        """Can get/set per-account mock run duration."""
        get_default = client.get("/api/v1/config/mock-run-duration/mock-runner")
        assert get_default.status_code == 200
        assert get_default.json()["seconds"] == 0

        set_resp = client.put(
            "/api/v1/config/mock-run-duration",
            json={"account_name": "mock-runner", "seconds": 60},
        )
        assert set_resp.status_code == 200
        assert set_resp.json()["seconds"] == 60

        get_after = client.get("/api/v1/config/mock-run-duration/mock-runner")
        assert get_after.status_code == 200
        assert get_after.json()["seconds"] == 60


class TestWorktreeEndpoints:
    """Tests for worktree endpoints (require valid session)."""

    def test_get_worktree_no_session(self, client):
        """Returns 404 for worktree on non-existent session."""
        response = client.get("/api/v1/sessions/nonexistent/worktree")
        assert response.status_code == 404

    def test_get_worktree_no_project(self, client):
        """Returns appropriate status when session has no worktree."""
        # Create session without project path
        create_resp = client.post("/api/v1/sessions", json={"name": "Test"})
        session_id = create_resp.json()["id"]

        response = client.get(f"/api/v1/sessions/{session_id}/worktree")
        assert response.status_code == 200
        data = response.json()
        assert data["exists"] is False

    def test_merge_request_accepts_commit_message(self, client, tmp_path, monkeypatch):
        """Merge request should accept optional commit_message field."""
        # Create a test git repo
        import subprocess
        project_dir = tmp_path / "project"
        project_dir.mkdir()
        subprocess.run(["git", "init"], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "config", "user.email", "test@test.com"], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "config", "user.name", "Test"], cwd=project_dir, capture_output=True)
        (project_dir / "file.txt").write_text("initial")
        subprocess.run(["git", "add", "."], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "commit", "-m", "initial"], cwd=project_dir, capture_output=True)

        # Create session with project path
        create_resp = client.post(
            "/api/v1/sessions",
            json={"name": "Test", "project_path": str(project_dir)},
        )
        session_id = create_resp.json()["id"]

        # Create worktree
        wt_resp = client.post(f"/api/v1/sessions/{session_id}/worktree")
        assert wt_resp.status_code == 201

        # Merge request with commit message should be accepted
        merge_resp = client.post(
            f"/api/v1/sessions/{session_id}/worktree/merge",
            json={"target_branch": None, "commit_message": "Custom merge commit"},
        )
        # Should not fail due to unknown field
        assert merge_resp.status_code in [200, 400]  # 400 is ok if no changes

    def test_get_branches_endpoint(self, client, tmp_path):
        """GET /worktree/branches should return branch list."""
        # Create a test git repo
        import subprocess
        project_dir = tmp_path / "project"
        project_dir.mkdir()
        subprocess.run(["git", "init"], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "config", "user.email", "test@test.com"], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "config", "user.name", "Test"], cwd=project_dir, capture_output=True)
        (project_dir / "file.txt").write_text("initial")
        subprocess.run(["git", "add", "."], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "commit", "-m", "initial"], cwd=project_dir, capture_output=True)

        # Create session with project path
        create_resp = client.post(
            "/api/v1/sessions",
            json={"name": "Test", "project_path": str(project_dir)},
        )
        session_id = create_resp.json()["id"]

        # Create worktree
        client.post(f"/api/v1/sessions/{session_id}/worktree")

        # Get branches
        response = client.get(f"/api/v1/sessions/{session_id}/worktree/branches")
        assert response.status_code == 200
        data = response.json()
        assert "branches" in data
        assert "default" in data
        assert "current" in data  # Current worktree branch
        assert isinstance(data["branches"], list)
        # Current branch is the worktree branch - it might be a chad-task-* branch
        # that isn't in the main branches list if it was just created for the worktree
        assert data["current"]  # Just verify it's not empty

    def test_resolve_conflicts_endpoint(self, client, tmp_path):
        """POST /worktree/resolve-conflicts should exist."""
        # Create a test git repo
        import subprocess
        project_dir = tmp_path / "project"
        project_dir.mkdir()
        subprocess.run(["git", "init"], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "config", "user.email", "test@test.com"], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "config", "user.name", "Test"], cwd=project_dir, capture_output=True)
        (project_dir / "file.txt").write_text("initial")
        subprocess.run(["git", "add", "."], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "commit", "-m", "initial"], cwd=project_dir, capture_output=True)

        # Create session with project path
        create_resp = client.post(
            "/api/v1/sessions",
            json={"name": "Test", "project_path": str(project_dir)},
        )
        session_id = create_resp.json()["id"]

        # Create worktree
        client.post(f"/api/v1/sessions/{session_id}/worktree")

        # Test endpoint exists (even without actual conflicts it should respond)
        response = client.post(
            f"/api/v1/sessions/{session_id}/worktree/resolve-conflicts",
            json={"use_incoming": True},
        )
        # Should return 200 or 400, not 404 (endpoint must exist)
        assert response.status_code != 404

    def test_abort_merge_endpoint(self, client, tmp_path):
        """POST /worktree/abort-merge should exist."""
        # Create a test git repo
        import subprocess
        project_dir = tmp_path / "project"
        project_dir.mkdir()
        subprocess.run(["git", "init"], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "config", "user.email", "test@test.com"], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "config", "user.name", "Test"], cwd=project_dir, capture_output=True)
        (project_dir / "file.txt").write_text("initial")
        subprocess.run(["git", "add", "."], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "commit", "-m", "initial"], cwd=project_dir, capture_output=True)

        # Create session with project path
        create_resp = client.post(
            "/api/v1/sessions",
            json={"name": "Test", "project_path": str(project_dir)},
        )
        session_id = create_resp.json()["id"]

        # Create worktree
        client.post(f"/api/v1/sessions/{session_id}/worktree")

        # Test endpoint exists
        response = client.post(f"/api/v1/sessions/{session_id}/worktree/abort-merge")
        # Should return 200 or 400, not 404 (endpoint must exist)
        assert response.status_code != 404

    def test_merge_cleans_up_session_state(self, client, tmp_path):
        """Successful merge should clear session worktree state."""
        import subprocess
        project_dir = tmp_path / "project"
        project_dir.mkdir()
        subprocess.run(["git", "init"], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "config", "user.email", "test@test.com"], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "config", "user.name", "Test"], cwd=project_dir, capture_output=True)
        (project_dir / "file.txt").write_text("initial")
        subprocess.run(["git", "add", "."], cwd=project_dir, capture_output=True)
        subprocess.run(["git", "commit", "-m", "initial"], cwd=project_dir, capture_output=True)

        # Create session with project path
        create_resp = client.post(
            "/api/v1/sessions",
            json={"name": "Test", "project_path": str(project_dir)},
        )
        session_id = create_resp.json()["id"]

        # Create worktree
        wt_resp = client.post(f"/api/v1/sessions/{session_id}/worktree")
        assert wt_resp.status_code == 201
        wt_data = wt_resp.json()
        worktree_path = wt_data["path"]

        # Make a change in the worktree
        import os
        (tmp_path / "project" / ".chad-worktrees").mkdir(exist_ok=True)
        if os.path.exists(worktree_path):
            with open(os.path.join(worktree_path, "new_file.txt"), "w") as f:
                f.write("new content")
            subprocess.run(["git", "add", "."], cwd=worktree_path, capture_output=True)
            subprocess.run(["git", "commit", "-m", "change"], cwd=worktree_path, capture_output=True)

        # Merge
        merge_resp = client.post(
            f"/api/v1/sessions/{session_id}/worktree/merge",
            json={"target_branch": None},
        )

        # After successful merge, worktree should not exist
        if merge_resp.status_code == 200 and merge_resp.json().get("success"):
            wt_status = client.get(f"/api/v1/sessions/{session_id}/worktree")
            assert wt_status.status_code == 200
            assert wt_status.json()["exists"] is False


class TestProjectSettingsEndpoints:
    """Tests for project settings API endpoints."""

    def test_get_project_settings_no_project(self, client):
        """Returns 400 when no project path provided."""
        response = client.get("/api/v1/config/project")
        assert response.status_code == 400

    def test_get_project_settings_nonexistent(self, client, tmp_path):
        """Returns default settings for new project."""
        project_path = str(tmp_path / "new-project")
        response = client.get(f"/api/v1/config/project?project_path={project_path}")
        assert response.status_code == 200
        data = response.json()
        assert data["project_path"] == project_path
        assert data["lint_command"] is None
        assert data["test_command"] is None

    def test_set_project_settings(self, client, tmp_path):
        """Can save project settings."""
        project_dir = tmp_path / "test-project"
        project_dir.mkdir()
        project_path = str(project_dir)

        response = client.put(
            "/api/v1/config/project",
            json={
                "project_path": project_path,
                "lint_command": "flake8 .",
                "test_command": "pytest tests/",
                "instructions_path": "AGENTS.md",
            },
        )
        assert response.status_code == 200
        data = response.json()
        assert data["lint_command"] == "flake8 ."
        assert data["test_command"] == "pytest tests/"
        assert data["instructions_path"] == "AGENTS.md"

    def test_get_project_settings_after_save(self, client, tmp_path):
        """Saved project settings are returned on GET."""
        project_dir = tmp_path / "test-project2"
        project_dir.mkdir()
        project_path = str(project_dir)

        # Save settings
        client.put(
            "/api/v1/config/project",
            json={
                "project_path": project_path,
                "lint_command": "npm run lint",
                "test_command": "npm test",
            },
        )

        # Retrieve settings
        response = client.get(f"/api/v1/config/project?project_path={project_path}")
        assert response.status_code == 200
        data = response.json()
        assert data["lint_command"] == "npm run lint"
        assert data["test_command"] == "npm test"

    def test_get_session_log_path(self, client):
        """Can get session log file path."""
        # Create a session
        create_resp = client.post("/api/v1/sessions", json={"name": "Test"})
        session_id = create_resp.json()["id"]

        response = client.get(f"/api/v1/sessions/{session_id}/log")
        assert response.status_code == 200
        data = response.json()
        assert "log_path" in data
        assert "log_exists" in data
        assert data["session_id"] == session_id
        # Log file may not exist until a task is started
        # If log_path is set, it should contain the session id
        if data["log_path"]:
            assert session_id in data["log_path"]


class TestConnectionLogging:
    """Tests for connection logging on SSE/WebSocket endpoints."""

    def test_websocket_logs_connection(self, client, capsys):
        """WebSocket endpoint should log connection events."""
        # Create a session first
        create_resp = client.post("/api/v1/sessions", json={"name": "Test WS"})
        session_id = create_resp.json()["id"]

        # Connect to WebSocket
        with client.websocket_connect(f"/api/v1/ws/{session_id}") as websocket:
            # Send a ping to establish connection
            websocket.send_json({"type": "ping"})
            response = websocket.receive_json()
            assert response["type"] == "pong"

        # Check that connection was logged
        captured = capsys.readouterr()
        assert f"WebSocket client connected to session {session_id}" in captured.out

    def test_websocket_logs_disconnection(self, client, capsys):
        """WebSocket endpoint should log disconnection events."""
        # Create a session first
        create_resp = client.post("/api/v1/sessions", json={"name": "Test Disconnect"})
        session_id = create_resp.json()["id"]

        # Connect and disconnect
        with client.websocket_connect(f"/api/v1/ws/{session_id}") as websocket:
            websocket.send_json({"type": "ping"})
            websocket.receive_json()
        # websocket is now disconnected

        # Check that disconnection was logged
        captured = capsys.readouterr()
        assert f"WebSocket client disconnected from session {session_id}" in captured.out


class TestHistoricalSessionEvents:
    """Tests for viewing historical events from finished sessions."""

    def test_events_endpoint_returns_historical_events_from_disk(self, client, tmp_path, monkeypatch):
        """Events endpoint should return events from persisted JSONL even without active task."""
        from chad.util.event_log import (
            EventLog,
            SessionStartedEvent,
            TerminalOutputEvent,
            SessionEndedEvent,
        )

        # Set up log directory
        log_dir = tmp_path / "logs"
        log_dir.mkdir()
        monkeypatch.setenv("CHAD_LOG_DIR", str(log_dir))

        # Create a session
        create_resp = client.post("/api/v1/sessions", json={"name": "Historical Test"})
        assert create_resp.status_code == 201
        session_id = create_resp.json()["id"]

        # Manually create a persisted event log (simulating a finished session)
        event_log = EventLog(session_id, base_dir=log_dir)
        event_log.log(SessionStartedEvent(
            task_description="Test task",
            project_path="/test/path",
            coding_provider="mock",
            coding_account="test-account",
        ))
        event_log.log(TerminalOutputEvent(data="Hello from terminal"))
        event_log.log(SessionEndedEvent(
            success=True,
            reason="completed",
            total_tool_calls=5,
            total_turns=3,
        ))
        event_log.close()

        # Now query the events endpoint - should return the persisted events
        response = client.get(f"/api/v1/sessions/{session_id}/events")
        assert response.status_code == 200
        data = response.json()

        # Should have events from the persisted log
        assert len(data["events"]) == 3
        assert data["latest_seq"] == 3

        # Check event types are present
        event_types = [e["type"] for e in data["events"]]
        assert "session_started" in event_types
        assert "terminal_output" in event_types
        assert "session_ended" in event_types

        # Check terminal output content
        terminal_events = [e for e in data["events"] if e["type"] == "terminal_output"]
        assert len(terminal_events) == 1
        assert terminal_events[0]["data"] == "Hello from terminal"

    def test_events_endpoint_filters_by_type_for_historical(self, client, tmp_path, monkeypatch):
        """Events endpoint should support filtering by type for historical events."""
        from chad.util.event_log import (
            EventLog,
            SessionStartedEvent,
            TerminalOutputEvent,
            SessionEndedEvent,
        )

        # Set up log directory
        log_dir = tmp_path / "logs"
        log_dir.mkdir()
        monkeypatch.setenv("CHAD_LOG_DIR", str(log_dir))

        # Create a session
        create_resp = client.post("/api/v1/sessions", json={"name": "Filter Test"})
        session_id = create_resp.json()["id"]

        # Create persisted event log
        event_log = EventLog(session_id, base_dir=log_dir)
        event_log.log(SessionStartedEvent(
            task_description="Test task",
            project_path="/test/path",
            coding_provider="mock",
            coding_account="test-account",
        ))
        event_log.log(TerminalOutputEvent(data="Line 1"))
        event_log.log(TerminalOutputEvent(data="Line 2"))
        event_log.log(SessionEndedEvent(success=True, reason="done"))
        event_log.close()

        # Filter to only terminal_output events
        response = client.get(
            f"/api/v1/sessions/{session_id}/events",
            params={"event_types": "terminal_output"},
        )
        assert response.status_code == 200
        data = response.json()

        assert len(data["events"]) == 2
        for event in data["events"]:
            assert event["type"] == "terminal_output"

    def test_events_endpoint_with_since_seq_for_historical(self, client, tmp_path, monkeypatch):
        """Events endpoint should support since_seq filtering for historical events."""
        from chad.util.event_log import EventLog, TerminalOutputEvent

        # Set up log directory
        log_dir = tmp_path / "logs"
        log_dir.mkdir()
        monkeypatch.setenv("CHAD_LOG_DIR", str(log_dir))

        # Create a session
        create_resp = client.post("/api/v1/sessions", json={"name": "Seq Test"})
        session_id = create_resp.json()["id"]

        # Create persisted event log with multiple events
        event_log = EventLog(session_id, base_dir=log_dir)
        for i in range(5):
            event_log.log(TerminalOutputEvent(data=f"Event {i}"))
        event_log.close()

        # Get events after seq 2
        response = client.get(
            f"/api/v1/sessions/{session_id}/events",
            params={"since_seq": 2},
        )
        assert response.status_code == 200
        data = response.json()

        assert len(data["events"]) == 3  # seqs 3, 4, 5
        seqs = [e["seq"] for e in data["events"]]
        assert seqs == [3, 4, 5]


class TestUploadEndpoint:
    """Tests for file upload functionality."""

    def test_upload_screenshot(self, client, tmp_path):
        """Can upload a screenshot file."""
        # Create a test image file (PNG-like bytes)
        png_header = b'\x89PNG\r\n\x1a\n' + b'\x00' * 100
        files = {"file": ("test.png", png_header, "image/png")}

        response = client.post("/api/v1/uploads", files=files)
        assert response.status_code == 201
        data = response.json()
        assert "path" in data
        assert data["filename"] == "test.png"
        # Verify the file was saved
        assert Path(data["path"]).exists()

    def test_upload_multiple_screenshots(self, client, tmp_path):
        """Can upload multiple screenshot files."""
        png_header = b'\x89PNG\r\n\x1a\n' + b'\x00' * 100

        # Upload first file
        response1 = client.post(
            "/api/v1/uploads",
            files={"file": ("screenshot1.png", png_header, "image/png")}
        )
        assert response1.status_code == 201

        # Upload second file
        response2 = client.post(
            "/api/v1/uploads",
            files={"file": ("screenshot2.png", png_header, "image/png")}
        )
        assert response2.status_code == 201

        # Paths should be different
        assert response1.json()["path"] != response2.json()["path"]

    def test_upload_rejects_non_image(self, client):
        """Rejects non-image file types."""
        response = client.post(
            "/api/v1/uploads",
            files={"file": ("malware.exe", b"bad content", "application/octet-stream")}
        )
        assert response.status_code == 400
        assert "image" in response.json()["detail"].lower()

    def test_upload_accepts_jpeg(self, client):
        """Accepts JPEG images."""
        jpeg_header = b'\xff\xd8\xff\xe0' + b'\x00' * 100
        response = client.post(
            "/api/v1/uploads",
            files={"file": ("photo.jpg", jpeg_header, "image/jpeg")}
        )
        assert response.status_code == 201

    def test_upload_accepts_webp(self, client):
        """Accepts WebP images."""
        webp_header = b'RIFF\x00\x00\x00\x00WEBP' + b'\x00' * 100
        response = client.post(
            "/api/v1/uploads",
            files={"file": ("image.webp", webp_header, "image/webp")}
        )
        assert response.status_code == 201



"""FastAPI application factory for Chad server."""

from contextlib import asynccontextmanager
from importlib import resources
from pathlib import Path
from typing import AsyncGenerator

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from . import __version__
from .state import init_start_time
from .api.routes import health, sessions, providers, worktree, config, ws, slack, tunnel, uploads


def _resolve_ui_index() -> Path | None:
    """Return the path to the UI index.html if available.

    The UI is a single self-contained HTML file (portable build) that can be
    served without any additional asset files.
    """
    # Prefer packaged assets (bundled in wheel)
    try:
        package_dist = resources.files("chad.ui_dist")
        index = Path(package_dist) / "index.html"
        if index.is_file():
            return index
    except Exception:
        pass

    # Fallback to portable build in repository (useful in editable installs)
    repo_portable = Path(__file__).resolve().parents[3] / "ui" / "dist-portable" / "index.html"
    if repo_portable.is_file():
        return repo_portable
    return None


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    """Application lifespan context manager."""
    init_start_time()

    # Startup: initialize services
    # TODO: Initialize session manager, etc.

    yield

    # Shutdown: cleanup resources
    # TODO: Cleanup sessions, stop providers, etc.


def create_app(
    title: str = "Chad Server",
    debug: bool = False,
    cors_origins: list[str] | None = None,
    auth_token: str | None = None,
) -> FastAPI:
    """Create and configure the FastAPI application.

    Args:
        title: Application title for OpenAPI docs
        debug: Enable debug mode
        cors_origins: List of allowed CORS origins (None = allow all)
        auth_token: Bearer token for API authentication (None = no auth)

    Returns:
        Configured FastAPI application
    """
    app = FastAPI(
        title=title,
        description="Backend API for Chad AI - multi-provider coding assistant",
        version=__version__,
        debug=debug,
        lifespan=lifespan,
    )

    # Store auth token on app state for WebSocket auth
    app.state.auth_token = auth_token

    # Configure CORS
    if cors_origins is None:
        # Default: allow all origins for development
        cors_origins = ["*"]

    # Add auth middleware before CORS so it runs after CORS (middleware order is LIFO)
    if auth_token:
        from .auth import BearerAuthMiddleware
        app.add_middleware(BearerAuthMiddleware, token=auth_token)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=cors_origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    # Include routers
    app.include_router(health.router, tags=["Health"])
    app.include_router(sessions.router, prefix="/api/v1/sessions", tags=["Sessions"])
    app.include_router(providers.router, prefix="/api/v1", tags=["Providers"])
    app.include_router(worktree.router, prefix="/api/v1/sessions", tags=["Worktree"])
    app.include_router(config.router, prefix="/api/v1/config", tags=["Config"])
    app.include_router(ws.router, prefix="/api/v1", tags=["WebSocket"])
    app.include_router(slack.router, prefix="/api/v1", tags=["Slack"])
    app.include_router(tunnel.router, prefix="/api/v1", tags=["Tunnel"])
    app.include_router(uploads.router, prefix="/api/v1/uploads", tags=["Uploads"])

    # Serve the single-file React UI if available (packaged or repo build).
    ui_index = _resolve_ui_index()
    if ui_index:
        from fastapi.responses import FileResponse

        @app.get("/", include_in_schema=False)
        async def serve_index():
            return FileResponse(ui_index)

    return app


# Create default application instance
app = create_app()



"""Model discovery and normalization for provider dropdowns."""

from __future__ import annotations

import base64
import json
import os
import shutil
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Iterable

from chad.util.utils import platform_path, safe_home

try:  # Python 3.11+
    import tomllib  # type: ignore
except ModuleNotFoundError:  # pragma: no cover
    try:
        import tomli as tomllib  # type: ignore
    except ModuleNotFoundError:
        tomllib = None  # type: ignore


def _safe_stat_mtime(path: Path) -> float:
    try:
        return path.stat().st_mtime
    except OSError:
        return 0.0


@dataclass
class ModelCatalog:
    """Discover and cache available models per provider."""

    api_client: object | None = None
    home_dir: Path = field(default_factory=safe_home)
    cache_ttl: float = 300.0
    max_session_files: int = 60

    # Only include "default" in fallback - specific models vary by account type
    # (ChatGPT accounts vs API accounts have different available models)
    # User's actual available models are discovered from config/session files
    OPENAI_FALLBACK: tuple[str, ...] = ("default",)
    ANTHROPIC_FALLBACK: tuple[str, ...] = (
        "claude-sonnet-4-20250514",
        "claude-opus-4-20250514",
        "default",
    )
    GEMINI_FALLBACK: tuple[str, ...] = (
        "gemini-2.5-pro",
        "gemini-2.5-flash",
        "gemini-2.5-flash-lite",
        "default",
    )
    QWEN_FALLBACK: tuple[str, ...] = (
        "qwen3-coder",
        "default",
    )
    MISTRAL_FALLBACK: tuple[str, ...] = ("default",)
    OPENCODE_FALLBACK: tuple[str, ...] = (
        "anthropic/claude-sonnet-4-5",
        "openai/gpt-4o",
        "default",
    )
    KIMI_FALLBACK: tuple[str, ...] = (
        "kimi-k2.5",
        "default",
    )
    MOCK_FALLBACK: tuple[str, ...] = ("default",)

    _cache: dict[str, tuple[float, list[str]]] = field(default_factory=dict, init=False)

    def supported_providers(self) -> set[str]:
        return {"anthropic", "openai", "gemini", "qwen", "mistral", "opencode", "kimi", "mock"}

    def get_models(self, provider: str, account_name: str | None = None) -> list[str]:
        """Return discovered models for a provider, cached with TTL."""
        cache_key = f"{provider}:{account_name or ''}"
        now = time.time()
        cached = self._cache.get(cache_key)
        if cached and now - cached[0] < self.cache_ttl:
            return cached[1]

        models = set(self._fallback(provider))
        models |= self._stored_model(provider, account_name)

        if provider == "openai":
            models |= self._codex_config_models(account_name)
            models |= self._codex_session_models(account_name)

        models = {str(m).strip() for m in models if m}
        models = {m for m in models if m}

        if provider == "openai":
            models = {m for m in models if m.lower() != "codex"}

        models.add("default")
        resolved = sorted(models, key=lambda m: (m == "default", m))
        self._cache[cache_key] = (now, resolved)
        return resolved

    # Discovery helpers -------------------------------------------------
    def _fallback(self, provider: str) -> Iterable[str]:
        return {
            "anthropic": self.ANTHROPIC_FALLBACK,
            "openai": self.OPENAI_FALLBACK,
            "gemini": self.GEMINI_FALLBACK,
            "qwen": self.QWEN_FALLBACK,
            "mistral": self.MISTRAL_FALLBACK,
            "opencode": self.OPENCODE_FALLBACK,
            "kimi": self.KIMI_FALLBACK,
            "mock": self.MOCK_FALLBACK,
        }.get(provider, ("default",))

    @staticmethod
    def _normalize_stored_model(model: object) -> str | None:
        if not isinstance(model, str):
            return None
        normalized = model.strip()
        return normalized or None

    def _model_matches_provider(self, provider: str, model: str) -> bool:
        normalized = model.strip().lower()
        if not normalized:
            return False
        if normalized == "default":
            return True

        if provider == "mock":
            return False
        if provider == "anthropic":
            return normalized.startswith("claude-")
        if provider == "gemini":
            return normalized.startswith("gemini-")
        if provider == "qwen":
            return normalized.startswith("qwen")
        if provider == "kimi":
            return normalized.startswith("kimi")
        if provider == "mistral":
            return normalized.startswith(("mistral", "codestral", "ministral", "open-mistral", "pixtral"))
        if provider == "opencode":
            return "/" in normalized
        if provider == "openai":
            foreign_prefixes = (
                "claude-",
                "anthropic/",
                "gemini-",
                "google/",
                "qwen",
                "kimi",
                "mistral",
                "codestral",
                "ministral",
                "open-mistral",
                "pixtral",
            )
            return not normalized.startswith(foreign_prefixes)
        return True

    def _stored_model(self, provider: str, account_name: str | None) -> set[str]:
        if not account_name or not self.api_client:
            return set()

        models: set[str] = set()

        # Primary: model stored on the account object (used throughout the UI)
        try:
            account = self.api_client.get_account(account_name)
            account_provider = getattr(account, "provider", None)
            if account_provider and str(account_provider).strip() != provider:
                return set()
            model = self._normalize_stored_model(getattr(account, "model", None))
            if model and self._model_matches_provider(provider, model):
                models.add(model)
        except Exception:
            pass

        # Legacy/compat: some API clients expose get_account_model instead
        try:
            model = self._normalize_stored_model(self.api_client.get_account_model(account_name))
            if model and self._model_matches_provider(provider, model):
                models.add(model)
        except Exception:
            pass

        return models

    def _codex_config_models(self, account_name: str | None) -> set[str]:
        if tomllib is None:
            return set()

        if not account_name:
            return set()

        self._sync_codex_home(account_name)
        config_path = self._codex_home(account_name) / ".codex" / "config.toml"
        if not config_path.exists():
            return set()

        try:
            data = tomllib.loads(config_path.read_text(encoding="utf-8"))
        except Exception:
            return set()

        models: set[str] = set()
        model = data.get("model")
        if model:
            models.add(str(model))

        notice = data.get("notice", {})
        if isinstance(notice, dict):
            migrations = notice.get("model_migrations", {})
            if isinstance(migrations, dict):
                for old, new in migrations.items():
                    if old:
                        models.add(str(old))
                    if new:
                        models.add(str(new))

        return models

    def _codex_session_models(self, account_name: str | None) -> set[str]:
        if not account_name:
            return set()

        self._sync_codex_home(account_name)
        sessions_dir = self._codex_home(account_name) / ".codex" / "sessions"
        if not sessions_dir.exists():
            return set()

        files = list(sessions_dir.rglob("*.jsonl"))
        files.sort(key=_safe_stat_mtime, reverse=True)
        models: set[str] = set()

        for path in files[: self.max_session_files]:
            try:
                with path.open(encoding="utf-8") as fh:
                    for line in fh:
                        line = line.strip()
                        if not line:
                            continue
                        try:
                            record = json.loads(line)
                        except json.JSONDecodeError:
                            continue

                        model = self._extract_model(record)
                        if model:
                            models.add(model)
            except OSError:
                continue

        return models

    @staticmethod
    def _extract_model(record: dict) -> str | None:
        direct = record.get("model")
        if direct:
            return str(direct)

        payload = record.get("payload")
        if isinstance(payload, dict):
            payload_model = payload.get("model")
            if payload_model:
                return str(payload_model)

        return None

    def _codex_home(self, account_name: str) -> Path:
        temp_home = os.environ.get("CHAD_TEMP_HOME")
        base = platform_path(temp_home) if temp_home else platform_path(self.home_dir)
        return platform_path(base / ".chad" / "codex-homes" / account_name)

    def _sync_codex_home(self, account_name: str) -> None:
        """Sync real-home Codex data into the isolated home.

        IMPORTANT: This only syncs files that DON'T already exist in the isolated home.
        Once an account has its own auth.json, it should never be overwritten by the
        real home's auth - that would cause multiple accounts to share credentials.
        """
        isolated_home = platform_path(self._codex_home(account_name) / ".codex")
        real_home = platform_path(safe_home(ignore_temp_home=True) / ".codex")
        if not real_home.exists():
            return

        isolated_home.mkdir(parents=True, exist_ok=True)

        def sync_file_if_missing(src: Path, dest: Path) -> None:
            """Only copy if destination doesn't exist - never overwrite existing auth."""
            try:
                if not dest.exists():
                    dest.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy2(src, dest)
            except OSError:
                return

        def sync_tree_if_missing(src_dir: Path, dest_dir: Path) -> None:
            """Sync tree but never overwrite existing files."""
            for root, _, files in os.walk(src_dir):
                for filename in files:
                    src_path = platform_path(root) / filename
                    rel_path = src_path.relative_to(src_dir)
                    dest_path = dest_dir / rel_path
                    sync_file_if_missing(src_path, dest_path)

        # Only sync auth.json if the isolated home doesn't have one yet
        # This prevents overwriting account-specific credentials
        sync_file_if_missing(real_home / "auth.json", isolated_home / "auth.json")
        sync_file_if_missing(real_home / "config.toml", isolated_home / "config.toml")
        if (real_home / "sessions").exists():
            sync_tree_if_missing(real_home / "sessions", isolated_home / "sessions")

    def _codex_plan_type(self, account_name: str | None) -> str | None:
        if not account_name:
            return None

        auth_file = self._codex_home(account_name) / ".codex" / "auth.json"
        if not auth_file.exists():
            return None

        try:
            data = json.loads(auth_file.read_text(encoding="utf-8"))
            token = data.get("tokens", {}).get("access_token")
            if not token:
                return None

            payload = token.split(".")[1]
            if payload:
                payload += "=" * (-len(payload) % 4)
            decoded = base64.urlsafe_b64decode(payload or "")
            jwt_data = json.loads(decoded)
            auth_info = jwt_data.get("https://api.openai.com/auth", {})
            plan_type = auth_info.get("chatgpt_plan_type")
            return str(plan_type).lower() if plan_type else None
        except Exception:
            return None



"""Pipe-based streaming service for Windows (no PTY support)."""

from __future__ import annotations

import asyncio
import base64
import os
import queue
import subprocess
import sys
import threading
import uuid
from dataclasses import dataclass, field
from pathlib import Path
from typing import AsyncIterator, Callable


@dataclass
class PTYSession:
    """Represents an active streaming session (pipe-based on Windows)."""

    stream_id: str
    session_id: str
    pid: int
    master_fd: int  # -1 on Windows (no PTY fd)
    cmd: list[str]
    cwd: Path
    env: dict[str, str]

    # Subprocess object
    _proc: subprocess.Popen | None = None
    _stdin_pipe: bool = False

    # State
    active: bool = True
    exit_code: int | None = None

    _subscribers: list[queue.Queue] = field(default_factory=list)
    _lock: threading.Lock = field(default_factory=threading.Lock)
    _event_buffer: list["PTYEvent"] = field(default_factory=list)
    _log_callback: "Callable[[PTYEvent], None] | None" = None
    _output_thread: threading.Thread | None = None


@dataclass
class PTYEvent:
    """An event from a streaming session."""

    type: str  # "output", "exit", "error"
    stream_id: str
    data: str = ""  # Base64 encoded for output
    exit_code: int | None = None
    error: str | None = None
    has_ansi: bool = True
    text: bool = False


class PTYStreamService:
    """Pipe-based stream service for Windows."""

    def __init__(self):
        self._sessions: dict[str, PTYSession] = {}
        self._lock = threading.Lock()

    def start_pty_session(
        self,
        session_id: str,
        cmd: list[str],
        cwd: Path,
        env: dict[str, str] | None = None,
        rows: int = 24,
        cols: int = 80,
        log_callback: Callable[["PTYEvent"], None] | None = None,
        stdin_pipe: bool = False,
    ) -> str:
        """Start a pipe-based process (Windows equivalent of PTY).

        Args:
            session_id: The session this belongs to
            cmd: Command and arguments to run
            cwd: Working directory
            env: Additional environment variables
            rows: Terminal rows (used for env vars only, no real terminal)
            cols: Terminal columns (used for env vars only, no real terminal)
            log_callback: Optional callback for logging events
            stdin_pipe: Ignored on Windows (always uses pipes)

        Returns:
            stream_id for this session
        """
        stream_id = f"pty_{uuid.uuid4().hex[:8]}"

        full_env = os.environ.copy()
        if env:
            full_env.update(env)

        full_env["LINES"] = str(rows)
        full_env["COLUMNS"] = str(cols)

        creation_flags = 0
        startupinfo = None
        if sys.platform == "win32":
            creation_flags = subprocess.CREATE_NEW_PROCESS_GROUP
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESTDHANDLES

        proc = subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=str(cwd),
            env=full_env,
            creationflags=creation_flags,
            startupinfo=startupinfo,
        )

        session = PTYSession(
            stream_id=stream_id,
            session_id=session_id,
            pid=proc.pid,
            master_fd=-1,
            cmd=cmd,
            cwd=cwd,
            env=full_env,
            _proc=proc,
            _stdin_pipe=True,
            _log_callback=log_callback,
        )

        with self._lock:
            self._sessions[stream_id] = session

        thread = threading.Thread(
            target=self._read_output_loop,
            args=(session,),
            daemon=True,
        )
        session._output_thread = thread
        thread.start()
        return stream_id

    def _read_output_loop(self, session: PTYSession) -> None:
        """Read output from process stdout pipe and dispatch to subscribers."""
        proc = session._proc
        if not proc or not proc.stdout:
            return

        while session.active:
            try:
                data = proc.stdout.read(4096)
                if data:
                    data = self._handle_cpr_request(session, data)
                    if not data:
                        continue
                    has_ansi = b"\x1b[" in data or b"\x1b]" in data
                    event = PTYEvent(
                        type="output",
                        stream_id=session.stream_id,
                        data=base64.b64encode(data).decode("ascii"),
                        has_ansi=has_ansi,
                    )
                    self._dispatch_event(session, event)
                else:
                    break
            except (OSError, ValueError):
                break

        # Get exit code
        if proc is not None:
            try:
                proc.wait(timeout=1.0)
                session.exit_code = proc.returncode
            except subprocess.TimeoutExpired:
                session.exit_code = -1

        event = PTYEvent(
            type="exit",
            stream_id=session.stream_id,
            exit_code=session.exit_code,
        )
        self._dispatch_event(session, event)
        session.active = False

    def _dispatch_event(self, session: PTYSession, event: PTYEvent) -> None:
        """Send event to logging callback and all subscribers."""
        if session._log_callback:
            try:
                session._log_callback(event)
            except Exception:
                pass

        with session._lock:
            session._event_buffer.append(event)
            if len(session._event_buffer) > 1000:
                session._event_buffer = session._event_buffer[-1000:]

            for q in session._subscribers:
                try:
                    q.put_nowait(event)
                except queue.Full:
                    pass

    def send_input(self, stream_id: str, data: bytes, close_stdin: bool = False) -> bool:
        """Send input to process stdin pipe."""
        with self._lock:
            session = self._sessions.get(stream_id)

        if not session or not session.active:
            return False

        try:
            if session._proc and session._proc.stdin:
                session._proc.stdin.write(data)
                session._proc.stdin.flush()
                if close_stdin:
                    session._proc.stdin.close()
            return True
        except OSError:
            return False

    def resize(self, stream_id: str, rows: int, cols: int) -> bool:
        """Resize terminal — no-op on Windows (pipes have no terminal size)."""
        return False

    async def subscribe(self, stream_id: str) -> AsyncIterator[PTYEvent]:
        """Subscribe to output events."""
        with self._lock:
            session = self._sessions.get(stream_id)

        if not session:
            return

        q: queue.Queue[PTYEvent] = queue.Queue(maxsize=1000)

        with session._lock:
            session._subscribers.append(q)
            buffered_events = list(session._event_buffer)

        for event in buffered_events:
            yield event
            if event.type == "exit":
                with session._lock:
                    if q in session._subscribers:
                        session._subscribers.remove(q)
                return

        try:
            while True:
                try:
                    event = q.get_nowait()
                    yield event
                    if event.type == "exit":
                        break
                except queue.Empty:
                    if not session.active:
                        while True:
                            try:
                                event = q.get_nowait()
                                yield event
                                if event.type == "exit":
                                    break
                            except queue.Empty:
                                break
                        break
                    await asyncio.sleep(0.05)
        finally:
            with session._lock:
                if q in session._subscribers:
                    session._subscribers.remove(q)

    def terminate(self, stream_id: str) -> bool:
        """Terminate a session."""
        with self._lock:
            session = self._sessions.get(stream_id)

        if not session:
            return False

        session.active = False

        if session._proc is not None:
            pid = session._proc.pid
            if sys.platform == "win32":
                try:
                    subprocess.run(
                        ["taskkill", "/F", "/T", "/PID", str(pid)],
                        capture_output=True,
                        timeout=5,
                    )
                except (subprocess.SubprocessError, OSError):
                    try:
                        session._proc.kill()
                    except OSError:
                        pass
            else:
                try:
                    session._proc.terminate()
                except OSError:
                    pass

                def force_kill():
                    try:
                        session._proc.kill()
                    except OSError:
                        pass

                threading.Timer(2.0, force_kill).start()

        return True

    def _handle_cpr_request(self, session: PTYSession, data: bytes) -> bytes:
        """Detect and respond to cursor position requests (CSI 6n)."""
        cpr_seq = b"\x1b[6n"
        if cpr_seq not in data:
            return data

        # Respond via stdin pipe
        try:
            if session._proc and session._proc.stdin:
                session._proc.stdin.write(b"\x1b[1;1R\r")
                session._proc.stdin.flush()
        except OSError:
            pass

        return data.replace(cpr_seq, b"")

    def get_session(self, stream_id: str) -> PTYSession | None:
        """Get a session by stream ID."""
        with self._lock:
            return self._sessions.get(stream_id)

    def get_session_by_session_id(self, session_id: str) -> PTYSession | None:
        """Get a session by the parent session ID."""
        with self._lock:
            matching = [s for s in self._sessions.values() if s.session_id == session_id]

        if not matching:
            return None

        for session in reversed(matching):
            if session.active:
                return session

        return matching[-1]

    def list_sessions(self) -> list[str]:
        """List all active stream IDs."""
        with self._lock:
            return [s.stream_id for s in self._sessions.values() if s.active]

    def cleanup_session(self, stream_id: str) -> None:
        """Remove a session from tracking."""
        with self._lock:
            if stream_id in self._sessions:
                del self._sessions[stream_id]

    def set_log_callback(
        self,
        stream_id: str,
        callback: Callable[["PTYEvent"], None] | None,
    ) -> bool:
        """Set the logging callback for a session."""
        with self._lock:
            session = self._sessions.get(stream_id)
            if not session:
                return False
            session._log_callback = callback
            return True


_pty_stream_service: PTYStreamService | None = None


def get_pty_stream_service() -> PTYStreamService:
    """Get the global PTYStreamService instance."""
    global _pty_stream_service
    if _pty_stream_service is None:
        _pty_stream_service = PTYStreamService()
    return _pty_stream_service


def reset_pty_stream_service() -> None:
    """Reset the global service (for testing)."""
    global _pty_stream_service
    if _pty_stream_service:
        for stream_id in _pty_stream_service.list_sessions():
            _pty_stream_service.terminate(stream_id)
    _pty_stream_service = None



"""Tests for message converter module."""

import tempfile
from pathlib import Path

import pytest

from chad.util.event_log import (
    EventLog,
    UserMessageEvent,
    AssistantMessageEvent,
    SessionStartedEvent,
)
from chad.util.message_converter import (
    ConversationTurn,
    extract_conversation_from_events,
    format_for_provider,
)


@pytest.fixture
def temp_log_dir():
    """Create a temporary directory for event logs."""
    with tempfile.TemporaryDirectory() as tmpdir:
        yield Path(tmpdir)


@pytest.fixture
def event_log(temp_log_dir):
    """Create an EventLog instance for testing."""
    return EventLog("test-session", base_dir=temp_log_dir)


class TestConversationTurn:
    """Tests for ConversationTurn dataclass."""

    def test_from_user_message(self):
        """Test creating user turn from text."""
        turn = ConversationTurn.from_user_message("Hello world", "2024-01-01T00:00:00Z")
        assert turn.role == "user"
        assert len(turn.blocks) == 1
        assert turn.blocks[0]["kind"] == "text"
        assert turn.blocks[0]["content"] == "Hello world"
        assert turn.timestamp == "2024-01-01T00:00:00Z"

    def test_from_assistant_blocks(self):
        """Test creating assistant turn from blocks."""
        blocks = [
            {"kind": "thinking", "content": "Let me analyze..."},
            {"kind": "text", "content": "Here is the answer."},
        ]
        turn = ConversationTurn.from_assistant_blocks(blocks)
        assert turn.role == "assistant"
        assert turn.blocks == blocks
        assert turn.timestamp is None


class TestExtractConversationFromEvents:
    """Tests for extract_conversation_from_events function."""

    def test_empty_log(self, event_log):
        """Test extraction from empty log returns empty list."""
        turns = extract_conversation_from_events(event_log)
        assert turns == []

    def test_extract_user_message(self, event_log):
        """Test extracting user message events."""
        event_log.log(UserMessageEvent(content="Add a button"))
        turns = extract_conversation_from_events(event_log)

        assert len(turns) == 1
        assert turns[0].role == "user"
        assert turns[0].blocks[0]["content"] == "Add a button"

    def test_extract_assistant_message(self, event_log):
        """Test extracting assistant message events."""
        event_log.log(
            AssistantMessageEvent(
                blocks=[
                    {"kind": "text", "content": "I'll help you add a button."},
                ]
            )
        )
        turns = extract_conversation_from_events(event_log)

        assert len(turns) == 1
        assert turns[0].role == "assistant"
        assert turns[0].blocks[0]["kind"] == "text"

    def test_extract_conversation_order(self, event_log):
        """Test that turns are extracted in chronological order."""
        event_log.log(UserMessageEvent(content="First message"))
        event_log.log(
            AssistantMessageEvent(blocks=[{"kind": "text", "content": "Response 1"}])
        )
        event_log.log(UserMessageEvent(content="Second message"))
        event_log.log(
            AssistantMessageEvent(blocks=[{"kind": "text", "content": "Response 2"}])
        )

        turns = extract_conversation_from_events(event_log)

        assert len(turns) == 4
        assert turns[0].role == "user"
        assert turns[0].blocks[0]["content"] == "First message"
        assert turns[1].role == "assistant"
        assert turns[2].role == "user"
        assert turns[2].blocks[0]["content"] == "Second message"
        assert turns[3].role == "assistant"

    def test_extract_with_since_seq(self, event_log):
        """Test filtering by sequence number."""
        event_log.log(UserMessageEvent(content="Old message"))
        seq = event_log.get_latest_seq()
        event_log.log(UserMessageEvent(content="New message"))

        turns = extract_conversation_from_events(event_log, since_seq=seq)

        assert len(turns) == 1
        assert turns[0].blocks[0]["content"] == "New message"

    def test_extract_with_max_turns(self, event_log):
        """Test limiting number of turns."""
        for i in range(5):
            event_log.log(UserMessageEvent(content=f"Message {i}"))

        turns = extract_conversation_from_events(event_log, max_turns=3)

        assert len(turns) == 3
        # Should get the last 3
        assert turns[0].blocks[0]["content"] == "Message 2"
        assert turns[2].blocks[0]["content"] == "Message 4"

    def test_ignores_other_events(self, event_log):
        """Test that non-message events are ignored."""
        event_log.log(SessionStartedEvent(task_description="Test task"))
        event_log.log(UserMessageEvent(content="Hello"))

        turns = extract_conversation_from_events(event_log)

        assert len(turns) == 1
        assert turns[0].role == "user"

    def test_assistant_with_tool_calls(self, event_log):
        """Test extracting assistant messages with tool calls."""
        event_log.log(
            AssistantMessageEvent(
                blocks=[
                    {"kind": "thinking", "content": "I need to read the file..."},
                    {
                        "kind": "tool_call",
                        "tool": "Read",
                        "args": {"file_path": "/src/main.py"},
                    },
                    {"kind": "tool_result", "content": "def main(): pass"},
                    {"kind": "text", "content": "I found the main function."},
                ]
            )
        )

        turns = extract_conversation_from_events(event_log)

        assert len(turns) == 1
        assert turns[0].role == "assistant"
        assert len(turns[0].blocks) == 4
        assert turns[0].blocks[0]["kind"] == "thinking"
        assert turns[0].blocks[1]["kind"] == "tool_call"


class TestFormatForProvider:
    """Tests for format_for_provider function."""

    def test_format_for_claude_omits_thinking(self):
        """Test that Claude format omits thinking blocks."""
        turns = [
            ConversationTurn.from_user_message("Help me"),
            ConversationTurn(
                role="assistant",
                blocks=[
                    {"kind": "thinking", "content": "Let me think..."},
                    {"kind": "text", "content": "Here is my answer."},
                ],
            ),
        ]

        result = format_for_provider(turns, "anthropic")

        assert "[User]: Help me" in result
        assert "Let me think" not in result
        assert "Here is my answer." in result

    def test_format_for_codex_includes_reasoning(self):
        """Test that Codex format includes reasoning blocks."""
        turns = [
            ConversationTurn.from_user_message("Help me"),
            ConversationTurn(
                role="assistant",
                blocks=[
                    {"kind": "thinking", "content": "Let me think..."},
                    {"kind": "text", "content": "Here is my answer."},
                ],
            ),
        ]

        result = format_for_provider(turns, "openai")

        assert "[User]: Help me" in result
        assert "[Reasoning]: Let me think..." in result
        assert "Here is my answer." in result

    def test_format_for_generic_uses_xml(self):
        """Test that generic format uses XML tags."""
        turns = [
            ConversationTurn.from_user_message("Help me"),
            ConversationTurn(
                role="assistant",
                blocks=[
                    {"kind": "thinking", "content": "Let me think..."},
                    {"kind": "text", "content": "Here is my answer."},
                ],
            ),
        ]

        result = format_for_provider(turns, "gemini")

        assert '<turn role="user">Help me</turn>' in result
        assert "<thinking>Let me think...</thinking>" in result
        assert "<response>Here is my answer.</response>" in result

    def test_format_tool_calls(self):
        """Test formatting of tool calls."""
        turns = [
            ConversationTurn(
                role="assistant",
                blocks=[
                    {
                        "kind": "tool_call",
                        "tool": "Read",
                        "args": {"file_path": "/src/main.py"},
                    },
                    {"kind": "tool_result", "content": "def main(): pass"},
                ],
            ),
        ]

        result = format_for_provider(turns, "openai")

        assert "[Tool: Read] /src/main.py" in result
        assert "[Result]: def main(): pass" in result

    def test_format_bash_command_truncation(self):
        """Test that long bash commands are truncated."""
        long_cmd = "echo " + "a" * 200
        turns = [
            ConversationTurn(
                role="assistant",
                blocks=[
                    {"kind": "tool_call", "tool": "Bash", "args": {"command": long_cmd}},
                ],
            ),
        ]

        result = format_for_provider(turns, "anthropic")

        assert "..." in result
        assert len(long_cmd) > 80  # Original is long
        # The formatted version should be truncated

    def test_format_with_new_message(self):
        """Test appending new message."""
        turns = [ConversationTurn.from_user_message("Previous")]

        result = format_for_provider(turns, "anthropic", new_message="New request")

        assert "[User]: Previous" in result
        assert "[User]: New request" in result

    def test_format_with_new_message_generic(self):
        """Test appending new message for generic provider."""
        turns = [ConversationTurn.from_user_message("Previous")]

        result = format_for_provider(turns, "gemini", new_message="New request")

        assert '<turn role="user">Previous</turn>' in result
        assert '<turn role="user">New request</turn>' in result

    def test_format_empty_turns(self):
        """Test formatting empty turns list."""
        result = format_for_provider([], "anthropic")
        assert result == ""

    def test_format_empty_turns_with_new_message(self):
        """Test formatting empty turns with new message."""
        result = format_for_provider([], "anthropic", new_message="Hello")
        assert "[User]: Hello" in result

    def test_preserves_full_tool_results(self):
        """Test that long tool results are preserved without truncation."""
        long_result = "x" * 1000
        turns = [
            ConversationTurn(
                role="assistant",
                blocks=[
                    {"kind": "tool_result", "content": long_result},
                ],
            ),
        ]

        for provider in ("anthropic", "openai", "gemini"):
            result = format_for_provider(turns, provider)
            assert long_result in result, f"Tool result truncated for {provider}"

    def test_preserves_full_thinking_blocks(self):
        """Test that long thinking blocks are preserved without truncation."""
        long_thinking = "y" * 2000
        turns = [
            ConversationTurn(
                role="assistant",
                blocks=[
                    {"kind": "thinking", "content": long_thinking},
                ],
            ),
        ]

        # Codex includes reasoning
        result = format_for_provider(turns, "openai")
        assert long_thinking in result

        # Generic includes thinking
        result = format_for_provider(turns, "gemini")
        assert long_thinking in result


class TestProviderSpecificFormatting:
    """Tests for provider-specific formatting edge cases."""

    def test_qwen_uses_generic_format(self):
        """Test that Qwen uses generic format."""
        turns = [ConversationTurn.from_user_message("Test")]
        result = format_for_provider(turns, "qwen")
        assert '<turn role="user">' in result

    def test_mistral_uses_generic_format(self):
        """Test that Mistral uses generic format."""
        turns = [ConversationTurn.from_user_message("Test")]
        result = format_for_provider(turns, "mistral")
        assert '<turn role="user">' in result

    def test_unknown_provider_uses_generic(self):
        """Test that unknown providers use generic format."""
        turns = [ConversationTurn.from_user_message("Test")]
        result = format_for_provider(turns, "unknown_provider")
        assert '<turn role="user">' in result

    def test_format_multiple_text_blocks(self):
        """Test formatting assistant with multiple text blocks."""
        turns = [
            ConversationTurn(
                role="assistant",
                blocks=[
                    {"kind": "text", "content": "First part."},
                    {"kind": "text", "content": "Second part."},
                ],
            ),
        ]

        result = format_for_provider(turns, "anthropic")

        assert "First part." in result
        assert "Second part." in result

    def test_format_glob_tool(self):
        """Test formatting Glob tool calls."""
        turns = [
            ConversationTurn(
                role="assistant",
                blocks=[
                    {
                        "kind": "tool_call",
                        "tool": "Glob",
                        "args": {"pattern": "**/*.py"},
                    },
                ],
            ),
        ]

        result = format_for_provider(turns, "openai")

        assert "[Tool: Glob] **/*.py" in result

    def test_format_grep_tool(self):
        """Test formatting Grep tool calls."""
        turns = [
            ConversationTurn(
                role="assistant",
                blocks=[
                    {
                        "kind": "tool_call",
                        "tool": "Grep",
                        "args": {"pattern": "def main"},
                    },
                ],
            ),
        ]

        result = format_for_provider(turns, "anthropic")

        assert "[Tool: Grep] def main" in result

    def test_format_websearch_tool(self):
        """Test formatting WebSearch tool calls."""
        turns = [
            ConversationTurn(
                role="assistant",
                blocks=[
                    {
                        "kind": "tool_call",
                        "tool": "WebSearch",
                        "args": {"query": "Python async tutorial"},
                    },
                ],
            ),
        ]

        result = format_for_provider(turns, "openai")

        assert "[Tool: WebSearch] Python async tutorial" in result



"""Unified streaming client for SSE and WebSocket connections."""

from __future__ import annotations

import base64
import json
import re
from dataclasses import dataclass
from typing import Any, AsyncIterator, Iterator

import httpx


@dataclass
class StreamEvent:
    """An event from the stream."""

    event_type: str  # terminal, event, ping, complete, error
    data: dict[str, Any]
    seq: int | None = None


class StreamClient:
    """Client for streaming API endpoints (SSE and input)."""

    def __init__(self, base_url: str = "http://localhost:3184"):
        """Initialize the stream client.

        Args:
            base_url: Base URL of the Chad server
        """
        self.base_url = base_url.rstrip("/")
        self._async_client: httpx.AsyncClient | None = None

    async def _get_async_client(self) -> httpx.AsyncClient:
        """Get or create async client."""
        if self._async_client is None:
            self._async_client = httpx.AsyncClient(timeout=None)
        return self._async_client

    async def close(self):
        """Close the client."""
        if self._async_client:
            await self._async_client.aclose()
            self._async_client = None

    def _url(self, path: str) -> str:
        """Build full URL for API path."""
        return f"{self.base_url}/api/v1{path}"

    async def stream_events(
        self,
        session_id: str,
        since_seq: int = 0,
        include_terminal: bool = True,
    ) -> AsyncIterator[StreamEvent]:
        """Stream events from a session via SSE.

        Args:
            session_id: Session to stream from
            since_seq: Resume from this sequence number
            include_terminal: Include raw PTY output

        Yields:
            StreamEvent objects
        """
        client = await self._get_async_client()
        url = self._url(f"/sessions/{session_id}/stream")
        params = {
            "since_seq": since_seq,
            "include_terminal": str(include_terminal).lower(),
        }

        async with client.stream("GET", url, params=params) as response:
            response.raise_for_status()

            buffer = ""
            current_event = ""
            current_data = ""

            async for chunk in response.aiter_text():
                buffer += chunk

                while "\n" in buffer:
                    line, buffer = buffer.split("\n", 1)
                    line = line.strip()

                    if line.startswith("event:"):
                        current_event = line[6:].strip()
                    elif line.startswith("data:"):
                        current_data = line[5:].strip()
                    elif line.startswith("id:"):
                        # Event ID
                        pass
                    elif line == "" and current_event and current_data:
                        # End of event
                        try:
                            data = json.loads(current_data)
                            seq = data.get("seq")
                            yield StreamEvent(
                                event_type=current_event,
                                data=data,
                                seq=seq,
                            )
                        except json.JSONDecodeError:
                            pass

                        current_event = ""
                        current_data = ""

    async def send_input(self, session_id: str, data: bytes) -> bool:
        """Send input to the PTY session.

        Args:
            session_id: Session ID
            data: Raw bytes to send

        Returns:
            True if sent successfully
        """
        client = await self._get_async_client()
        url = self._url(f"/sessions/{session_id}/input")
        encoded = base64.b64encode(data).decode("ascii")

        try:
            response = await client.post(url, json={"data": encoded})
            response.raise_for_status()
            return True
        except Exception:
            return False

    async def resize_terminal(
        self,
        session_id: str,
        rows: int,
        cols: int,
    ) -> bool:
        """Resize the PTY terminal.

        Args:
            session_id: Session ID
            rows: Number of rows
            cols: Number of columns

        Returns:
            True if resized successfully
        """
        client = await self._get_async_client()
        url = self._url(f"/sessions/{session_id}/resize")

        try:
            response = await client.post(url, json={"rows": rows, "cols": cols})
            response.raise_for_status()
            return True
        except Exception:
            return False


class SyncStreamClient:
    """Synchronous wrapper around StreamClient for non-async code."""

    def __init__(self, base_url: str = "http://localhost:3184"):
        """Initialize the sync stream client.

        Args:
            base_url: Base URL of the Chad server
        """
        self.base_url = base_url.rstrip("/")
        self._sync_client = httpx.Client(timeout=None)

    def close(self):
        """Close the client."""
        self._sync_client.close()

    def _url(self, path: str) -> str:
        """Build full URL for API path."""
        return f"{self.base_url}/api/v1{path}"

    def stream_events(
        self,
        session_id: str,
        since_seq: int = 0,
        include_terminal: bool = True,
    ) -> Iterator[StreamEvent]:
        """Stream events from a session via SSE (blocking).

        Args:
            session_id: Session to stream from
            since_seq: Resume from this sequence number
            include_terminal: Include raw PTY output

        Yields:
            StreamEvent objects
        """
        url = self._url(f"/sessions/{session_id}/stream")
        params = {
            "since_seq": since_seq,
            "include_terminal": str(include_terminal).lower(),
        }

        with self._sync_client.stream("GET", url, params=params) as response:
            response.raise_for_status()

            buffer = ""
            current_event = ""
            current_data = ""

            for chunk in response.iter_text():
                buffer += chunk

                while "\n" in buffer:
                    line, buffer = buffer.split("\n", 1)
                    line = line.strip()

                    if line.startswith("event:"):
                        current_event = line[6:].strip()
                    elif line.startswith("data:"):
                        current_data = line[5:].strip()
                    elif line.startswith("id:"):
                        pass
                    elif line == "" and current_event and current_data:
                        try:
                            data = json.loads(current_data)
                            seq = data.get("seq")
                            yield StreamEvent(
                                event_type=current_event,
                                data=data,
                                seq=seq,
                            )
                        except json.JSONDecodeError:
                            pass

                        current_event = ""
                        current_data = ""

    def send_input(self, session_id: str, data: bytes) -> bool:
        """Send input to the PTY session (blocking).

        Args:
            session_id: Session ID
            data: Raw bytes to send

        Returns:
            True if sent successfully
        """
        url = self._url(f"/sessions/{session_id}/input")
        encoded = base64.b64encode(data).decode("ascii")

        try:
            response = self._sync_client.post(url, json={"data": encoded})
            response.raise_for_status()
            return True
        except Exception:
            return False

    def resize_terminal(
        self,
        session_id: str,
        rows: int,
        cols: int,
    ) -> bool:
        """Resize the PTY terminal (blocking).

        Args:
            session_id: Session ID
            rows: Number of rows
            cols: Number of columns

        Returns:
            True if resized successfully
        """
        url = self._url(f"/sessions/{session_id}/resize")

        try:
            response = self._sync_client.post(url, json={"rows": rows, "cols": cols})
            response.raise_for_status()
            return True
        except Exception:
            return False


def decode_terminal_data(data: str | bytes, *, is_text: bool = False) -> bytes:
    """Decode terminal output data.

    Args:
        data: Terminal payload from an event
        is_text: True when payload is already plain text (not base64)

    Returns:
        Terminal bytes suitable for writing to a TTY
    """
    if is_text:
        if isinstance(data, bytes):
            result = data
        else:
            result = str(data or "").encode("utf-8", errors="replace")
    elif not data:
        return b""
    else:
        result = base64.b64decode(data)

    # Normalize bare LFs to CRLF so cursor returns to column 0 on new lines.
    # Some terminal emulators (e.g., in IDEs) treat LF as line-feed only,
    # causing text drift. The regex uses negative lookbehind to avoid double-CRLF.
    text = result.decode("utf-8", errors="replace")
    text = re.sub(r"(?<!\r)\n", "\r\n", text)
    return text.encode("utf-8", errors="replace")



"""Playwright helpers for launching the React UI and capturing screenshots.

Provides the minimal surface needed by scripts/release_screenshots.py and
any future visual tests:

    from chad.util.verification.ui_runner import (
        ChadLaunchError, PlaywrightUnavailable,
        create_temp_env, start_chad, stop_chad, open_playwright_page,
    )
"""

from __future__ import annotations

import base64
import contextlib
import os
import socket
import subprocess
import sys
import tempfile
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Iterator, Optional

import bcrypt

from chad.util.config_manager import ConfigManager
from chad.util.process_registry import ProcessRegistry

PROJECT_ROOT = Path(__file__).resolve().parents[4]  # src/chad/util/verification -> project root
SHARED_BROWSERS_PATH = Path.home() / ".cache" / "ms-playwright"

# ---------------------------------------------------------------------------
# Exceptions
# ---------------------------------------------------------------------------


class PlaywrightUnavailable(RuntimeError):
    """Raised when Playwright or Chromium are missing."""


class ChadLaunchError(RuntimeError):
    """Raised when the Chad server cannot be started or reached."""


# ---------------------------------------------------------------------------
# Data classes
# ---------------------------------------------------------------------------


@dataclass
class TempChadEnv:
    """Temporary environment for running Chad + Playwright."""

    config_path: Path
    project_dir: Path
    temp_dir: Path
    password: str = ""
    env_vars: dict = field(default_factory=dict)

    def cleanup(self) -> None:
        if os.environ.get("CHAD_CONFIG") == str(self.config_path):
            os.environ.pop("CHAD_CONFIG")
        try:
            import shutil
            shutil.rmtree(self.temp_dir, ignore_errors=True)
        except Exception:
            pass


@dataclass
class ChadInstance:
    """Running Chad process details."""

    process: subprocess.Popen[str]
    port: int
    env: TempChadEnv


# ---------------------------------------------------------------------------
# Registry (singleton per-process)
# ---------------------------------------------------------------------------

_test_registry: ProcessRegistry | None = None


def _get_test_registry() -> ProcessRegistry:
    global _test_registry
    if _test_registry is None:
        _test_registry = ProcessRegistry()
    return _test_registry


# ---------------------------------------------------------------------------
# Playwright helpers
# ---------------------------------------------------------------------------


def _has_chromium_cache(browsers_path: Path) -> bool:
    chromium_dirs = list(browsers_path.glob("chromium-*"))
    return any((d / "chrome-linux").is_dir() or (d / "chrome-win").is_dir() for d in chromium_dirs)


def ensure_playwright_browsers() -> None:
    browsers_path = SHARED_BROWSERS_PATH
    browsers_path.mkdir(parents=True, exist_ok=True)
    if _has_chromium_cache(browsers_path):
        return
    env = os.environ.copy()
    env["PLAYWRIGHT_BROWSERS_PATH"] = os.fspath(browsers_path)
    env["NODE_OPTIONS"] = env.get("NODE_OPTIONS", "") + " --dns-result-order=ipv4first"
    result = subprocess.run(
        [sys.executable, "-m", "playwright", "install", "chromium"],
        capture_output=True, text=True, encoding="utf-8", errors="replace", env=env,
    )
    if result.returncode != 0:
        detail = (result.stderr.strip() or result.stdout.strip())
        raise PlaywrightUnavailable(f"Failed to install Playwright browsers: {detail}")


def ensure_playwright():
    try:
        from playwright.sync_api import sync_playwright
        ensure_playwright_browsers()
        return sync_playwright
    except ImportError as exc:
        raise PlaywrightUnavailable(
            "Playwright is not installed. Install with `pip install playwright` "
            "and run `playwright install chromium`."
        ) from exc


# ---------------------------------------------------------------------------
# Temp environment
# ---------------------------------------------------------------------------


def create_temp_env(screenshot_mode: bool = True) -> TempChadEnv:
    """Create a temporary Chad config and project for UI testing."""
    temp_dir = Path(tempfile.mkdtemp(prefix="chad_ui_runner_"))
    config_path = temp_dir / "config.json"
    project_dir = temp_dir / "project"
    project_dir.mkdir(parents=True, exist_ok=True)
    (project_dir / "README.md").write_text("# Test Project\n")

    security_mgr = ConfigManager(config_path)
    password = ""
    password_hash = security_mgr.hash_password(password)
    encryption_salt = base64.urlsafe_b64encode(bcrypt.gensalt()).decode()

    config = {
        "password_hash": password_hash,
        "encryption_salt": encryption_salt,
        "accounts": {},
    }
    security_mgr.save_config(config)

    if screenshot_mode:
        # Create fixture accounts for release screenshots
        security_mgr.store_account(
            "claude-pro", "anthropic", "", password, "claude-sonnet-4-20250514"
        )
        security_mgr.store_account(
            "codex-work", "openai", "", password, "o3-pro"
        )
        security_mgr.assign_role("claude-pro", "CODING")
        # Configure action rules that showcase different actions
        security_mgr.set_action_settings([
            {"event": "session_usage", "threshold": 100, "action": "await_reset"},
            {
                "event": "weekly_usage",
                "threshold": 90,
                "action": "switch_provider",
                "target_account": "codex-work",
            },
            {"event": "context_usage", "threshold": 90, "action": "notify"},
        ])
        # Set a clean project path for the task form screenshot
        security_mgr.save_preferences("/home/user/my-webapp")
    else:
        security_mgr.store_account("mock-coding", "mock", "", password, "mock-model")
        security_mgr.assign_role("mock-coding", "CODING")

    return TempChadEnv(
        config_path=config_path,
        project_dir=project_dir,
        temp_dir=temp_dir,
        password=password,
        env_vars={"CHAD_SCREENSHOT_MODE": "1"} if screenshot_mode else {},
    )


# ---------------------------------------------------------------------------
# Server lifecycle
# ---------------------------------------------------------------------------


def _find_free_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind(("127.0.0.1", 0))
        return int(sock.getsockname()[1])


def _wait_for_ready(port: int, timeout: int = 30, process: subprocess.Popen[str] | None = None) -> None:
    """Wait until the API /status endpoint responds."""
    import urllib.request
    url = f"http://127.0.0.1:{port}/status"
    start = time.time()
    while time.time() - start < timeout:
        if process is not None and process.poll() is not None:
            raise ChadLaunchError("Chad server exited before becoming ready")
        try:
            urllib.request.urlopen(url, timeout=5)
            return
        except Exception:
            time.sleep(0.5)
    raise ChadLaunchError("Timed out waiting for Chad API to become ready")


def start_chad(env: TempChadEnv) -> ChadInstance:
    """Start Chad API server with an ephemeral port and return the running instance."""
    registry = _get_test_registry()
    registry.cleanup_stale()

    chad_env = {
        **os.environ,
        "CHAD_CONFIG": os.fspath(env.config_path),
        "CHAD_PASSWORD": env.password,
        "CHAD_PROJECT_PATH": os.fspath(env.project_dir),
        "PYTHONPATH": os.fspath(PROJECT_ROOT / "src"),
        "CHAD_PARENT_PID": str(os.getpid()),
    }
    if env.env_vars:
        chad_env.update(env.env_vars)

    popen_kwargs: Dict[str, object] = {}
    if os.name != "nt":
        popen_kwargs["start_new_session"] = True

    requested_port = _find_free_port()
    process = subprocess.Popen(
        [
            os.fspath(Path(sys.executable)),
            "-m", "chad",
            "--mode", "server",
            "--api-port", str(requested_port),
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        stdin=subprocess.DEVNULL,
        text=True, encoding="utf-8", errors="replace", bufsize=1,
        env=chad_env,
        cwd=os.fspath(PROJECT_ROOT),
        **popen_kwargs,
    )
    registry.register(process, description=f"chad test server (port {requested_port})")
    _wait_for_ready(requested_port, process=process)
    return ChadInstance(process=process, port=requested_port, env=env)


def stop_chad(instance: ChadInstance) -> None:
    """Terminate a running Chad instance."""
    registry = _get_test_registry()
    registry.terminate(instance.process.pid, timeout=5.0)


# ---------------------------------------------------------------------------
# Playwright page context
# ---------------------------------------------------------------------------


def _select_tab(page, tab: str) -> None:
    """Click a tab button in the React UI header nav."""
    buttons = page.locator("nav.tabs button")
    count = buttons.count()
    for i in range(count):
        btn = buttons.nth(i)
        if btn.inner_text().strip().lower() == tab.lower():
            btn.click()
            page.wait_for_timeout(500)
            return
    raise ChadLaunchError(f"Could not find tab matching '{tab}'")


@contextlib.contextmanager
def open_playwright_page(
    port: int,
    *,
    tab: Optional[str] = None,
    headless: bool = True,
    viewport: Optional[Dict[str, int]] = None,
    color_scheme: str | None = "dark",
    render_delay: float = 1.0,
) -> Iterator:
    """Open a Playwright page pointed at the React UI (served via Vite or static)."""
    sync_playwright = ensure_playwright()
    if viewport is None:
        viewport = {"width": 1280, "height": 900}

    with sync_playwright() as p:
        launch_args = ["--disable-gpu"]
        browser = p.chromium.launch(headless=headless, args=launch_args)
        context = browser.new_context(viewport=viewport, color_scheme=color_scheme)
        page = context.new_page()
        try:
            page.goto(f"http://127.0.0.1:{port}", wait_until="domcontentloaded", timeout=30000)
            # Wait for the React app to render
            page.wait_for_selector(".app-header h1", timeout=15000)
            if tab:
                _select_tab(page, tab)
            if render_delay > 0:
                page.wait_for_timeout(int(render_delay * 1000))
            yield page
        finally:
            context.close()
            browser.close()



"""Tests for project_setup module."""

import json
import os
import subprocess

import pytest

from chad.util.project_setup import (
    detect_project_type,
    detect_verification_commands,
    detect_python_executable,
    detect_doc_paths,
    ensure_docs_config,
    validate_command,
    load_project_config,
    save_project_config,
    save_project_settings,
    setup_project,
    build_verification_instructions,
    build_doc_reference_text,
    ProjectConfig,
    VerificationConfig,
)


@pytest.fixture
def isolated_config(tmp_path, monkeypatch):
    """Create an isolated config file for testing project settings."""
    config_path = tmp_path / "chad_test.conf"
    monkeypatch.setenv("CHAD_CONFIG", str(config_path))
    return config_path


class TestDetectProjectType:
    """Test cases for detect_project_type."""

    def test_detect_python_pyproject(self, tmp_path):
        """Test detection of Python project via pyproject.toml."""
        (tmp_path / "pyproject.toml").write_text("[project]\nname = 'test'\n")
        assert detect_project_type(tmp_path) == "python"

    def test_detect_python_setup_py(self, tmp_path):
        """Test detection of Python project via setup.py."""
        (tmp_path / "setup.py").write_text("from setuptools import setup\nsetup()")
        assert detect_project_type(tmp_path) == "python"

    def test_detect_python_requirements(self, tmp_path):
        """Test detection of Python project via requirements.txt."""
        (tmp_path / "requirements.txt").write_text("flask\nrequests\n")
        assert detect_project_type(tmp_path) == "python"

    def test_detect_javascript(self, tmp_path):
        """Test detection of JavaScript project via package.json."""
        (tmp_path / "package.json").write_text('{"name": "test"}')
        assert detect_project_type(tmp_path) == "javascript"

    def test_detect_typescript(self, tmp_path):
        """Test detection of TypeScript project via tsconfig.json."""
        (tmp_path / "tsconfig.json").write_text('{"compilerOptions": {}}')
        assert detect_project_type(tmp_path) == "typescript"

    def test_detect_rust(self, tmp_path):
        """Test detection of Rust project via Cargo.toml."""
        (tmp_path / "Cargo.toml").write_text("[package]\nname = 'test'\n")
        assert detect_project_type(tmp_path) == "rust"

    def test_detect_go(self, tmp_path):
        """Test detection of Go project via go.mod."""
        (tmp_path / "go.mod").write_text("module test\n")
        assert detect_project_type(tmp_path) == "go"

    def test_detect_unknown(self, tmp_path):
        """Test detection of unknown project type."""
        assert detect_project_type(tmp_path) == "unknown"


class TestDetectPythonExecutable:
    """Test cases for detect_python_executable."""

    def test_detect_venv_bin(self, tmp_path):
        """Test detection of .venv/bin/python."""
        venv_python = tmp_path / ".venv" / "bin" / "python"
        venv_python.parent.mkdir(parents=True)
        venv_python.write_text("#!/usr/bin/env python3")
        assert detect_python_executable(tmp_path) == str(venv_python)

    def test_detect_venv_scripts(self, tmp_path):
        """Test detection of .venv/Scripts/python.exe (Windows style)."""
        venv_python = tmp_path / ".venv" / "Scripts" / "python.exe"
        venv_python.parent.mkdir(parents=True)
        venv_python.write_text("stub")
        assert detect_python_executable(tmp_path) == str(venv_python)

    def test_fallback_to_python3(self, tmp_path):
        """Test fallback to python3 when no venv found."""
        assert detect_python_executable(tmp_path) == "python3"


class TestDetectVerificationCommands:
    """Test cases for detect_verification_commands."""

    def test_detect_python_commands(self, tmp_path):
        """Test detection of Python verification commands."""
        (tmp_path / "pyproject.toml").write_text("[project]\n")
        (tmp_path / "tests").mkdir()

        result = detect_verification_commands(tmp_path)
        assert result["project_type"] == "python"
        assert "flake8" in result["lint_command"]
        assert "pytest" in result["test_command"]

    def test_detect_makefile_commands(self, tmp_path):
        """Test detection of Makefile targets."""
        (tmp_path / "Makefile").write_text("lint:\n\tflake8 .\n\ntest:\n\tpytest\n")
        (tmp_path / "pyproject.toml").write_text("[project]\n")

        result = detect_verification_commands(tmp_path)
        assert result["lint_command"] == "make lint"
        assert result["test_command"] == "make test"

    def test_detect_javascript_commands(self, tmp_path):
        """Test detection of JavaScript verification commands."""
        package_json = {"name": "test", "scripts": {"lint": "eslint .", "test": "jest"}}
        (tmp_path / "package.json").write_text(json.dumps(package_json))

        result = detect_verification_commands(tmp_path)
        assert result["project_type"] == "javascript"
        assert result["lint_command"] == "npm run lint"
        assert result["test_command"] == "npm test"

    def test_javascript_missing_scripts(self, tmp_path):
        """Test JavaScript project without lint/test scripts."""
        package_json = {"name": "test"}  # No scripts
        (tmp_path / "package.json").write_text(json.dumps(package_json))

        result = detect_verification_commands(tmp_path)
        assert result["lint_command"] is None
        assert result["test_command"] is None


class TestValidateCommand:
    """Test cases for validate_command."""

    def test_validate_successful_command(self, tmp_path):
        """Test validation of a successful command."""
        success, output = validate_command("echo hello", tmp_path, timeout=5)
        assert success is True
        assert "hello" in output

    def test_validate_failing_command(self, tmp_path):
        """Test validation of a failing command."""
        success, output = validate_command("exit 1", tmp_path, timeout=5)
        assert success is False

    def test_validate_nonexistent_command(self, tmp_path):
        """Test validation of a non-existent command."""
        success, output = validate_command("nonexistent_command_xyz", tmp_path, timeout=5)
        assert success is False


class TestProjectConfig:
    """Test cases for ProjectConfig save/load."""

    def test_save_and_load_config(self, tmp_path, isolated_config):
        """Test saving and loading project configuration."""
        project_path = tmp_path / "my_project"
        project_path.mkdir()

        config = ProjectConfig(
            project_type="python",
            verification=VerificationConfig(
                lint_command="flake8 .",
                test_command="pytest tests/",
                validated=True,
            ),
        )

        save_project_config(project_path, config)

        # Verify config was saved to main config file
        assert isolated_config.exists()

        # Load and verify
        loaded = load_project_config(project_path)
        assert loaded is not None
        assert loaded.project_type == "python"
        assert loaded.verification.lint_command == "flake8 ."
        assert loaded.verification.test_command == "pytest tests/"
        assert loaded.verification.validated is True

    def test_load_nonexistent_config(self, tmp_path, isolated_config):
        """Test loading config when it doesn't exist."""
        result = load_project_config(tmp_path)
        assert result is None

    def test_config_to_dict(self):
        """Test ProjectConfig serialization to dict."""
        config = ProjectConfig(
            project_type="rust",
            verification=VerificationConfig(
                lint_command="cargo clippy",
                test_command="cargo test",
            ),
        )
        data = config.to_dict()

        assert data["project_type"] == "rust"
        assert data["verification"]["lint_command"] == "cargo clippy"
        assert data["verification"]["test_command"] == "cargo test"

    def test_config_from_dict(self):
        """Test ProjectConfig deserialization from dict."""
        data = {
            "version": "1.0",
            "project_type": "go",
            "verification": {
                "lint_command": "golint ./...",
                "test_command": "go test ./...",
                "lint_timeout": 30,
                "test_timeout": 120,
                "validated": True,
            },
        }
        config = ProjectConfig.from_dict(data)

        assert config.project_type == "go"
        assert config.verification.lint_command == "golint ./..."
        assert config.verification.validated is True


class TestSetupProject:
    """Test cases for setup_project."""

    def test_setup_new_project(self, tmp_path, isolated_config):
        """Test setting up a new project with detection."""
        (tmp_path / "pyproject.toml").write_text("[project]\n")
        (tmp_path / "tests").mkdir()

        # Run setup without validation (validation would try to run commands)
        config = setup_project(tmp_path, validate=False)

        assert config.project_type == "python"
        assert "flake8" in config.verification.lint_command
        assert "pytest" in config.verification.test_command

    def test_setup_loads_existing_config(self, tmp_path, isolated_config):
        """Test that setup loads existing validated config."""
        existing_config = ProjectConfig(
            project_type="custom",
            verification=VerificationConfig(
                lint_command="custom_lint",
                test_command="custom_test",
                validated=True,
            ),
        )
        save_project_config(tmp_path, existing_config)

        # Should return existing config
        config = setup_project(tmp_path, validate=False)
        assert config.verification.lint_command == "custom_lint"


class TestBuildVerificationInstructions:
    """Test cases for build_verification_instructions."""

    def test_instructions_with_validated_config(self, tmp_path, isolated_config):
        """Test building instructions from validated config."""
        config = ProjectConfig(
            project_type="python",
            verification=VerificationConfig(
                lint_command="flake8 src/",
                test_command="pytest tests/",
                validated=True,
            ),
        )
        save_project_config(tmp_path, config)

        instructions = build_verification_instructions(tmp_path)
        assert "flake8 src/" in instructions
        assert "pytest tests/" in instructions

    def test_instructions_with_auto_detection(self, tmp_path, isolated_config):
        """Test building instructions via auto-detection."""
        (tmp_path / "pyproject.toml").write_text("[project]\n")
        (tmp_path / "tests").mkdir()

        instructions = build_verification_instructions(tmp_path)
        assert "flake8" in instructions
        assert "pytest" in instructions

    def test_instructions_unknown_project(self, tmp_path, isolated_config):
        """Test building instructions for unknown project type."""
        instructions = build_verification_instructions(tmp_path)
        assert "verification patterns" in instructions.lower()
        assert "pytest" in instructions  # Generic hints


class TestDocsConfig:
    """Test detection and prompt reference generation for docs."""

    def test_detect_doc_paths_prefers_agents_and_architecture(self, tmp_path):
        """Detect doc paths and store them relative to project root."""
        (tmp_path / "AGENTS.md").write_text("instructions", encoding="utf-8")
        (tmp_path / "docs").mkdir()
        (tmp_path / "docs" / "ARCHITECTURE.md").write_text("arch", encoding="utf-8")

        docs = detect_doc_paths(tmp_path)
        assert docs.instructions_path == "AGENTS.md"
        assert docs.architecture_path == "docs/ARCHITECTURE.md"

    def test_ensure_docs_config_persists_to_config_file(self, tmp_path, isolated_config):
        """ensure_docs_config should save discovered paths into main config file."""
        (tmp_path / "AGENTS.md").write_text("instructions", encoding="utf-8")
        docs = ensure_docs_config(tmp_path)
        assert docs.instructions_path == "AGENTS.md"

        saved = load_project_config(tmp_path)
        assert saved is not None
        assert saved.docs.instructions_path == "AGENTS.md"

    def test_build_doc_reference_text_outputs_absolute_paths(self, tmp_path, isolated_config):
        """Doc reference text should point to absolute file locations."""
        (tmp_path / "AGENTS.md").write_text("instructions", encoding="utf-8")
        ref_text = build_doc_reference_text(tmp_path)
        assert ref_text is not None
        assert str(tmp_path / "AGENTS.md") in ref_text

    def test_save_project_settings_respects_custom_doc_paths(self, tmp_path, isolated_config):
        """Custom doc paths saved via project settings should be used in prompts."""
        docs_dir = tmp_path / "docs"
        docs_dir.mkdir()
        custom_instructions = docs_dir / "DEV_GUIDE.md"
        custom_architecture = tmp_path / "ARCH.md"
        custom_instructions.write_text("custom instructions", encoding="utf-8")
        custom_architecture.write_text("arch overview", encoding="utf-8")

        saved = save_project_settings(
            tmp_path,
            lint_command="flake8 .",
            test_command="pytest -q",
            instructions_path="docs/DEV_GUIDE.md",
            architecture_path="ARCH.md",
        )

        assert saved.docs.instructions_path == "docs/DEV_GUIDE.md"
        assert saved.docs.architecture_path == "ARCH.md"

        ref_text = build_doc_reference_text(tmp_path)
        assert str(custom_instructions.resolve()) in ref_text
        assert str(custom_architecture.resolve()) in ref_text

    def test_build_doc_reference_text_uses_saved_doc_paths_in_worktree(self, tmp_path, isolated_config):
        """Worktree prompt docs should use saved project doc paths, not auto-detected defaults."""
        from chad.util.git_worktree import GitWorktreeManager

        (tmp_path / "AGENTS.md").write_text("default instructions", encoding="utf-8")
        docs_dir = tmp_path / "docs"
        docs_dir.mkdir()
        custom_instructions = docs_dir / "DEV_GUIDE.md"
        custom_instructions.write_text("custom instructions", encoding="utf-8")
        (tmp_path / "README.md").write_text("repo", encoding="utf-8")

        subprocess.run(["git", "init"], cwd=tmp_path, check=True, capture_output=True, text=True)
        subprocess.run(["git", "add", "."], cwd=tmp_path, check=True, capture_output=True, text=True)
        commit_env = os.environ.copy()
        commit_env.update(
            {
                "GIT_AUTHOR_NAME": "test",
                "GIT_AUTHOR_EMAIL": "test@example.com",
                "GIT_COMMITTER_NAME": "test",
                "GIT_COMMITTER_EMAIL": "test@example.com",
            }
        )
        subprocess.run(
            ["git", "commit", "-m", "init"],
            cwd=tmp_path,
            check=True,
            capture_output=True,
            text=True,
            env=commit_env,
        )

        save_project_settings(
            tmp_path,
            lint_command="flake8 .",
            test_command="pytest -q",
            instructions_path="docs/DEV_GUIDE.md",
            architecture_path=None,
        )

        manager = GitWorktreeManager(tmp_path)
        worktree_path, _ = manager.create_worktree("docscheck")
        worktree_path = worktree_path.resolve()
        try:
            ref_text = build_doc_reference_text(worktree_path)
            assert ref_text is not None
            assert str(worktree_path / "docs" / "DEV_GUIDE.md") in ref_text
            assert str(worktree_path / "AGENTS.md") not in ref_text
        finally:
            subprocess.run(
                ["git", "worktree", "remove", str(worktree_path), "--force"],
                cwd=tmp_path,
                check=False,
                capture_output=True,
                text=True,
            )



"""Test cases for Qwen model catalog functionality."""

import os
from pathlib import Path

from chad.util.model_catalog import ModelCatalog


def test_qwen_fallback_models_do_not_include_unsupported_model(monkeypatch, tmp_path):
    """Test that qwen3-coder-plus is not included in Qwen fallback models."""
    monkeypatch.setenv("CHAD_TEMP_HOME", str(tmp_path))
    if os.name == "nt":
        monkeypatch.setattr(Path, "home", classmethod(lambda cls: tmp_path))

    catalog = ModelCatalog(home_dir=tmp_path)
    models = catalog.get_models("qwen", "test-account")

    # Verify that the unsupported model is not in the fallback models
    assert "qwen3-coder-plus" not in models, (
        "qwen3-coder-plus should not be in the fallback models as it's not supported by Qwen CLI"
    )
    
    # Verify that supported models are still present
    assert "qwen3-coder" in models
    assert "default" in models


def test_qwen_provider_specific_models(monkeypatch, tmp_path):
    """Test that Qwen provider returns appropriate models."""
    monkeypatch.setenv("CHAD_TEMP_HOME", str(tmp_path))
    if os.name == "nt":
        monkeypatch.setattr(Path, "home", classmethod(lambda cls: tmp_path))

    catalog = ModelCatalog(home_dir=tmp_path)
    
    # Get fallback models specifically
    fallback_models = catalog._fallback("qwen")
    
    # Verify the exact fallback models
    expected_fallbacks = {"qwen3-coder", "default"}
    actual_fallbacks = set(fallback_models)
    
    assert actual_fallbacks == expected_fallbacks, (
        f"Expected fallback models {expected_fallbacks}, got {actual_fallbacks}"
    )
    
    # Verify that unsupported model is not in fallback
    assert "qwen3-coder-plus" not in actual_fallbacks
